/*==========================================================================
 * Copyright (c) 2004 University of Massachusetts.  All Rights Reserved.
 *
 * Use of the Lemur Toolkit for Language Modeling and Information Retrieval
 * is subject to the terms of the software license set forth in the LICENSE
 * file included with this software, and also available at
 * http://www.lemurproject.org/license.html
 *
 *==========================================================================
 */

//
// FrequencyListCopier
//
// 24 August 2004 -- tds
//
// Finds IndexTerm nodes that only need to return frequency information,
// not positions, and inserts IndexFrequencyTerms instead.
//

#ifndef VANK_FREQUENCYLISTCOPIER_HPP
#define VANK_FREQUENCYLISTCOPIER_HPP

#include "ListCache.hpp"
namespace vank
{
  namespace lang
  {

    class FrequencyListCopier : public vank::lang::Copier {
    private:
      std::vector<vank::lang::Node*> _nodes;
      std::stack<vank::lang::Node*> _disqualifiers;
      vank::lang::IndexTerm* _lastTerm;
      bool _disqualifiedTree;

      ListCache* _listCache;

    public:
      FrequencyListCopier( ListCache* listCache ) : _listCache(listCache), _lastTerm(0), _disqualifiedTree(false) {}

      vank::lang::Node* defaultAfter( vank::lang::Node* oldNode, vank::lang::Node* newNode ) {
        if( _disqualifiers.size() && oldNode == _disqualifiers.top() )
          _disqualifiers.pop();

        _nodes.push_back( newNode );
        return newNode;
      }

      ~FrequencyListCopier() {
        vank::utility::delete_vector_contents<vank::lang::Node*>( _nodes );
      }

      void before( vank::lang::ExtentAnd* exAnd ) {
        _disqualifiedTree = true;
        _disqualifiers.push(exAnd);
      }

      void before( vank::lang::ExtentOr* exOr ) {
        _disqualifiedTree = true;
      }

      void before( vank::lang::ExtentInside* exInside ) {
        _disqualifiedTree = true;
      }

      void before( vank::lang::NestedExtentInside* nestExInside ) {
        _disqualifiedTree = true;
      }

      void before( vank::lang::ExtentRestriction* exRestrict ) {
        _disqualifiedTree = true;
        _disqualifiers.push(exRestrict);
      }

      void before( vank::lang::ExtentEnforcement* exEnforce ) {
        _disqualifiedTree = true;
        _disqualifiers.push(exEnforce);
      }

      void before( vank::lang::FixedPassage* fixedPassage ) {
        _disqualifiedTree = true;
        _disqualifiers.push(fixedPassage);
      }

      void before( vank::lang::FieldGreaterNode* fieldGreater ) {
        _disqualifiedTree = true;
        _disqualifiers.push(fieldGreater);
      }

      void before( vank::lang::FieldLessNode* fieldLess ) {
        _disqualifiedTree = true;
        _disqualifiers.push(fieldLess);
      }

      void before( vank::lang::FieldEqualsNode* fieldEquals ) {
        _disqualifiedTree = true;
        _disqualifiers.push(fieldEquals);
      }

      void before( vank::lang::FieldBetweenNode* fieldBetween ) {
        _disqualifiedTree = true;
        _disqualifiers.push(fieldBetween);
      }

      void before( vank::lang::ContextCounterNode* context ) {
        if( context->getContext() != NULL ) {
          _disqualifiedTree = true;
        }
      }

      void before( vank::lang::WeightedExtentOr* wExOr ) {
        _disqualifiedTree = true;
      }

      void before( vank::lang::ODNode* odNode ) {
        _disqualifiedTree = true;
      }

      void before( vank::lang::UWNode* uwNode ) {
        _disqualifiedTree = true;
      }

      void before( vank::lang::BAndNode* bandNode ) {
        _disqualifiedTree = true;
      }

      vank::lang::Node* after( vank::lang::IndexTerm* oldNode, vank::lang::IndexTerm* newNode ) {
        _lastTerm = newNode;
        return defaultAfter( oldNode, newNode );
      }

      void before( vank::lang::RawScorerNode* oldNode, vank::lang::RawScorerNode* newNode ) {
        _lastTerm = 0;
        _disqualifiedTree = false;
      }


      void before( vank::lang::NestedRawScorerNode* oldNode, vank::lang::NestedRawScorerNode* newNode ) {
        before( (vank::lang::RawScorerNode*) oldNode, (vank::lang::RawScorerNode*) newNode );
      }

      vank::lang::Node* after( vank::lang::RawScorerNode* oldNode, vank::lang::RawScorerNode* newNode ) {
        vank::lang::Node* result = 0;

        if( _lastTerm && !_disqualifiers.size() && !_disqualifiedTree && oldNode->getContext() == NULL ) {
          vank::lang::TermFrequencyScorerNode* scorerNode;
          // there's a term to score, and nothing to disqualify us from doing frequency scoring
          scorerNode = new vank::lang::TermFrequencyScorerNode( _lastTerm->getText(),
                                                                 _lastTerm->getStemmed() );

          scorerNode->setNodeName( oldNode->nodeName() );
          scorerNode->setSmoothing( oldNode->getSmoothing() );
          scorerNode->setStatistics( oldNode->getOccurrences(), oldNode->getContextSize(), oldNode->getDocumentOccurrences(), oldNode->getDocumentCount() );

          delete newNode;
          result = defaultAfter( oldNode, scorerNode );
        } else if( !_disqualifiers.size() ) {
          ListCache::CachedList* list = 0;

          if( _listCache )
            list = _listCache->find( newNode->getRawExtent(), newNode->getContext() );

          if( list ) {
            vank::lang::CachedFrequencyScorerNode* cachedNode;
            cachedNode = new vank::lang::CachedFrequencyScorerNode( newNode->getRawExtent(), newNode->getContext() );
            cachedNode->setNodeName( newNode->nodeName() );
            cachedNode->setSmoothing( newNode->getSmoothing() );
            cachedNode->setList( list );

            delete newNode;
            result = defaultAfter( oldNode, cachedNode );
          } else {
            result = defaultAfter( oldNode, newNode );
          }
        } else {
          result = defaultAfter( oldNode, newNode );
        }

        _disqualifiedTree = false;
        return result;
      }

      vank::lang::Node* after( vank::lang::NestedRawScorerNode* oldNode, vank::lang::NestedRawScorerNode* newNode ) {
        return after( (vank::lang::RawScorerNode*) oldNode, (vank::lang::RawScorerNode*) newNode );
      }
    };
  }
}

#endif // VANK_FREQUENCYLISTCOPIER_HPP
