/*==========================================================================
 * Copyright (c) 2004 University of Massachusetts.  All Rights Reserved.
 *
 * Use of the Lemur Toolkit for Language Modeling and Information Retrieval
 * is subject to the terms of the software license set forth in the LICENSE
 * file included with this software, and also available at
 * http://www.lemurproject.org/license.html
 *
 *==========================================================================
 */

//
// IndexWriter
//
// 26 November 2004 -- tds
//

#ifndef VANK_INDEXWRITER_HPP
#define VANK_INDEXWRITER_HPP

#include <vector>
#include <utility>
#include <queue>

#include "lemur/lemur-compat.hpp"
#include "vank/vank-platform.h"
#include "vank/greedy_vector"
#include "vank/TermData.hpp"
#include "lemur/Keyfile.hpp"
#include "vank/Index.hpp"
#include "vank/DocListFileIterator.hpp"
#include "vank/File.hpp"
#include "vank/SequentialWriteBuffer.hpp"
#include "vank/CorpusStatistics.hpp"
#include "vank/FieldStatistics.hpp"
#include "vank/TermBitmap.hpp"
#include "vank/TermRecorder.hpp"
#include "vank/TermTranslator.hpp"
#include "vank/DeletedDocumentList.hpp"
#include "vank/BulkTree.hpp"

namespace vank {
  namespace index {

    struct WriterIndexContext {
      struct greater {
      private:
        vank::index::DocListFileIterator::iterator_greater _iterator_greater;

        int _compareTerms( const WriterIndexContext* const&  one, const WriterIndexContext* const& two ) const {
          const char* oneTerm = one->iterator->currentEntry()->termData->term;
          const char* twoTerm = two->iterator->currentEntry()->termData->term;

          return strcmp( oneTerm, twoTerm );
        }

        int _compareDocuments( const WriterIndexContext* const&  one, const WriterIndexContext* const& two ) const {
          const vank::index::DocListIterator::DocumentData* oneData = one->iterator->currentEntry()->iterator->currentEntry();
          const vank::index::DocListIterator::DocumentData* twoData = two->iterator->currentEntry()->iterator->currentEntry();

          lemur::api::DOCID_T oneDocument = oneData ? oneData->document + one->documentOffset : 0;
          lemur::api::DOCID_T twoDocument = twoData ? twoData->document + two->documentOffset : 0;

          return oneDocument > twoDocument;
        }

      public:
        bool operator () ( const WriterIndexContext* const&  one, const WriterIndexContext* const& two ) const {
          assert( !one->iterator->finished() && !two->iterator->finished() );

          int result = _compareTerms( one, two );

          // if terms don't match, we're done
          if( result != 0 )
            return result > 0;

          // terms match, so go by document
          return _compareDocuments( one, two ) > 0;
        }
      };

      WriterIndexContext( vank::index::Index* _index, vank::index::DeletedDocumentList* _deletedList, lemur::api::DOCID_T _documentOffset ) {
        deletedList = _deletedList;
        documentOffset = _documentOffset;

        bitmap = new vank::index::TermBitmap;
        index = _index;
        wasInfrequentCount = 0;
        wasFrequentCount = 0;

        if( index->iteratorLock() )
          index->iteratorLock()->lock();

        iterator = index->docListFileIterator();
        iterator->startIteration();

        newlyFrequent = new vank::index::TermRecorder;
        oldFrequent = new vank::index::TermRecorder;
        oldInfrequent = new vank::utility::HashTable<lemur::api::TERMID_T, lemur::api::TERMID_T>;

        // DEBUG
        sequenceCount = 0;
      }

      ~WriterIndexContext() {
        delete iterator;

        if( index->iteratorLock() )
          index->iteratorLock()->unlock();

        delete oldFrequent;
        delete newlyFrequent;
        delete oldInfrequent;
        delete bitmap;
      }

      vank::index::DocListFileIterator* iterator;
      vank::index::TermBitmap* bitmap;
      vank::index::Index* index;

      int wasFrequentCount;
      int wasInfrequentCount;
      int sequenceCount;
      vank::index::TermRecorder* newlyFrequent;
      vank::index::TermRecorder* oldFrequent;
      vank::utility::HashTable<lemur::api::TERMID_T, lemur::api::TERMID_T>* oldInfrequent;

      vank::index::DeletedDocumentList* deletedList;
      lemur::api::DOCID_T documentOffset;
    };

    typedef std::priority_queue<WriterIndexContext*,
                                std::vector<WriterIndexContext*>,
                                WriterIndexContext::greater> invertedlist_pqueue;

    class IndexWriter {
    private:
      struct disktermdata_count_greater {
        bool operator () ( const DiskTermData* one, const DiskTermData* two ) const {
          return one->termData->corpus.totalCount > two->termData->corpus.totalCount;
        }
      };

      struct disktermdata_alpha_less {
        bool operator () ( const DiskTermData* one, const DiskTermData* two ) const {
          return strcmp( one->termData->term, two->termData->term ) < 0;
        }
      };

      struct keyfile_pair {
        vank::file::BulkTreeWriter* stringMap;
        vank::file::BulkTreeWriter* idMap;
      };

      keyfile_pair _infrequentTerms;
      keyfile_pair _frequentTerms;
      vank::file::File _frequentTermsData;

      vank::file::BulkTreeReader _infrequentTermsReader;
      vank::file::BulkTreeReader _frequentTermsReader;

      vank::file::File _documentStatistics;
      vank::file::File _documentLengths;

      vank::file::File _invertedFile;
      vank::file::File _directFile;
      vank::file::File _fieldsFile;

      vank::file::SequentialWriteBuffer* _invertedOutput;

      vank::utility::greedy_vector<vank::index::DiskTermData*> _topTerms;
      int _topTermsCount;
      vank::utility::Buffer _termDataBuffer;

      int _isFrequentCount;
      lemur::api::DOCID_T _documentBase;
      vank::index::CorpusStatistics _corpus;
      std::vector<vank::index::Index::FieldDescription> _fields;
      std::vector<vank::index::FieldStatistics> _fieldData;

      void _writeManifest( const std::string& path );
      void _writeSkip( vank::file::SequentialWriteBuffer* buffer, lemur::api::DOCID_T document, int length );
      void _writeBatch( vank::file::SequentialWriteBuffer* buffer, lemur::api::DOCID_T document, int length, vank::utility::Buffer& data );

      void _writeFieldLists( std::vector<WriterIndexContext*>& contexts, const std::string& path );
      void _writeFieldList( vank::file::SequentialWriteBuffer& output, int fieldIndex, std::vector<vank::index::DocExtentListIterator*>& iterators, std::vector<WriterIndexContext*>& contexts );

      void _pushInvertedLists( vank::utility::greedy_vector<WriterIndexContext*>& lists, invertedlist_pqueue& queue );
      void _fetchMatchingInvertedLists( vank::utility::greedy_vector<WriterIndexContext*>& lists, invertedlist_pqueue& queue );
      void _writeStatistics( vank::utility::greedy_vector<WriterIndexContext*>& lists, vank::index::TermData* termData, UINT64& startOffset );
      void _writeInvertedLists( std::vector<WriterIndexContext*>& contexts );

      void _storeIdEntry( IndexWriter::keyfile_pair& pair, vank::index::DiskTermData* diskTermData );
      void _storeStringEntry( IndexWriter::keyfile_pair& pair, vank::index::DiskTermData* diskTermData );

      void _storeTermEntry( IndexWriter::keyfile_pair& pair, vank::index::DiskTermData* diskTermData );
      void _storeFrequentTerms();
      void _addInvertedListData( vank::utility::greedy_vector<WriterIndexContext*>& lists, vank::index::TermData* termData, vank::utility::Buffer& listBuffer, UINT64& endOffset );
      void _storeMatchInformation( vank::utility::greedy_vector<WriterIndexContext*>& lists, int sequence, vank::index::TermData* termData, UINT64 startOffset, UINT64 endOffset );

      lemur::api::TERMID_T _lookupTermID( vank::file::BulkTreeReader& keyfile, const char* term );

      void _buildIndexContexts( std::vector<WriterIndexContext*>& contexts, std::vector<vank::index::Index*>& indexes, vank::index::DeletedDocumentList& deletedList );
      void _buildIndexContexts( std::vector<WriterIndexContext*>& contexts, std::vector<vank::index::Index*>& indexes, std::vector<vank::index::DeletedDocumentList*>& deletedLists, const std::vector<lemur::api::DOCID_T>& documentOffsets );

      void _writeDirectLists( std::vector<WriterIndexContext*>& contexts );
      void _writeDirectLists( WriterIndexContext* context,
                              vank::file::SequentialWriteBuffer* directOutput,
                              vank::file::SequentialWriteBuffer* lengthsOutput,
                              vank::file::SequentialWriteBuffer* dataOutput );

      void _constructFiles( const std::string& path );
      void _closeFiles( const std::string& path );
      void _openTermsReaders( const std::string& path );

      vank::index::TermTranslator* _buildTermTranslator( vank::file::BulkTreeReader& newInfrequentTerms,
                                                          vank::file::BulkTreeReader& newFrequentTerms,
                                                          vank::index::TermRecorder& oldFrequentTermsRecorder,
                                                          vank::utility::HashTable<lemur::api::TERMID_T, lemur::api::TERMID_T>* oldInfrequent,
                                                          vank::index::TermRecorder& newFrequentTermsRecorder,
                                                          vank::index::Index* index,
                                                          vank::index::TermBitmap* bitmap );

      // buffers for _lookupTermID
      char *_compressedData;
      char *_uncompressedData;
      int _dataSize;

      enum {
        TOPDOCS_DOCUMENT_COUNT = 1000,
        FREQUENT_TERM_COUNT = 1000
      };

    public:
      IndexWriter();
      void write( vank::index::Index& index,
                  std::vector<vank::index::Index::FieldDescription>& fields,
                  vank::index::DeletedDocumentList& deletedList,
                  const std::string& fileName );
      void write( std::vector<vank::index::Index*>& indexes,
                  std::vector<vank::index::Index::FieldDescription>& fields,
                  vank::index::DeletedDocumentList& deletedList,
                  const std::string& fileName );
      void write( std::vector<vank::index::Index*>& indexes,
                  std::vector<vank::index::Index::FieldDescription>& fields,
                  std::vector<vank::index::DeletedDocumentList*>& deletedLists,
                  const std::vector<lemur::api::DOCID_T>& documentMaximums,
                  const std::string& path );
    };
  }
}

#endif // VANK_INDEXWRITER_HPP
