/*==========================================================================
 * Copyright (c) 2004 University of Massachusetts.  All Rights Reserved.
 *
 * Use of the Lemur Toolkit for Language Modeling and Information Retrieval
 * is subject to the terms of the software license set forth in the LICENSE
 * file included with this software, and also available at
 * http://www.lemurproject.org/license.html
 *
 *==========================================================================
 */


//
// SmoothingAnnotatorWalker
//
// 27 April 2004 -- tds
//

#ifndef VANK_SMOOTHINGANNOTATORWALKER_HPP
#define VANK_SMOOTHINGANNOTATORWALKER_HPP

#include "vank/Parameters.hpp"
namespace vank
{
  namespace lang
  {

    class SmoothingAnnotatorWalker : public vank::lang::Walker {
    private:
      struct rule_type {
        std::string node;
        std::string field;
        std::string op;
        std::string smoothing;
      };

      std::vector<rule_type*> _rules;
      std::string _defaultSmoothing;

      void _loadSmoothingRules( vank::api::Parameters& parameters ) {
        if( !parameters.exists("rule") )
          return;

        vank::api::Parameters rules = parameters["rule"];

        for(size_t i=0; i<rules.size(); i++) {
          std::string ruleText = rules[i];

          int nextComma = 0;
          int nextColon = 0;
          int location = 0;

          rule_type* rule = new rule_type;
          rule->node = "RawScorerNode";
          rule->op = "*";
          rule->field = "*";

          for( location = 0; location < ruleText.length(); ) {
            nextComma = ruleText.find( ',', location );
            nextColon = ruleText.find( ':', location );

            std::string key = ruleText.substr( location, nextColon-location );
            std::string value = ruleText.substr( nextColon+1, nextComma-nextColon-1 );

            if( key == "node" ) {
              rule->node = value;
            } else if( key == "field" ) {
              rule->field = value;
            } else if( key == "operator" ) {
              rule->op = value;
            }  else {
              if( rule->smoothing.size() ) rule->smoothing += ",";
              rule->smoothing += key + ":" + value;
            }

            if( nextComma > 0 )
              location = nextComma+1;
            else
              location = ruleText.size();
          }

          _rules.push_back(rule);
        }
      }

      const std::string& _matchSmoothingRule( const std::string& node, const std::string& field, const std::string& op ) {
        for( int i=signed(_rules.size())-1; i >= 0; i-- ) {
          const rule_type& rule = *_rules[i];

          if( ( rule.node == node ) &&
              ( rule.field == field || rule.field == "*" ) &&
              ( rule.op == op || rule.op == "*" ) ) {
            return rule.smoothing;
          }
        }

        return _defaultSmoothing;
      }

    public:
      SmoothingAnnotatorWalker( vank::api::Parameters& parameters ) {
        _loadSmoothingRules( parameters );
        _defaultSmoothing = "method:dirichlet,mu:2500";
      }

      ~SmoothingAnnotatorWalker( ) {
        vank::utility::delete_vector_contents<rule_type*>( _rules );
      }

      void after( vank::lang::RawScorerNode* scorer ) {
        vank::lang::Node* context = scorer->getContext();
        vank::lang::Field* contextField = dynamic_cast<vank::lang::Field*>(context);
        vank::lang::ExtentOr* contextExtOr = dynamic_cast<vank::lang::ExtentOr*>(context);
        std::string fieldName;

        // there may be an ExtentOr around the field, so descend into it if necessary
        if( contextExtOr && contextExtOr->getChildren().size() == 1 ) {
          contextField = dynamic_cast<vank::lang::Field*>(contextExtOr->getChildren()[0]);
        }

        // if there's a field here, record its name
        if( contextField ) {
          fieldName = contextField->getFieldName();
        } else {
          fieldName = "?";
        }

        vank::lang::Node* raw = scorer->getRawExtent();
        vank::lang::Node* rawTerm = dynamic_cast<vank::lang::IndexTerm*>(raw);
        vank::lang::Node* rawODNode = dynamic_cast<vank::lang::ODNode*>(raw);
        vank::lang::Node* rawUWNode = dynamic_cast<vank::lang::UWNode*>(raw);
        vank::lang::Node* rawWeightedExtentOr = dynamic_cast<vank::lang::WeightedExtentOr*>(raw);

        std::string op;

        if( rawODNode || rawUWNode ) {
          op = "window";
        } else if( rawTerm || rawWeightedExtentOr ) {
          op = "term";
        } else {
          op = "?";
        }

        scorer->setSmoothing( _matchSmoothingRule( "RawScorerNode", fieldName, op ) );
      }

      void after( vank::lang::NestedRawScorerNode* scorer ) {
        after( (vank::lang::RawScorerNode *) scorer );
      }

      void after( vank::lang::ShrinkageScorerNode* scorer ) {
        after( (vank::lang::RawScorerNode *) scorer );

        for( int i=signed(_rules.size())-1; i >= 0; i-- ) {
          const rule_type& rule = *_rules[i];

          if( rule.node == "ShrinkageBelief" &&
              rule.op == "*" ) {
            if ( rule.field == "*" ) {
              scorer->addShrinkageRule( rule.smoothing );
            } else {
              std::string ruleString = "field:" + rule.field + "," + rule.smoothing;
              scorer->addShrinkageRule( ruleString );
            }
          }
        }
      }

      void after( vank::lang::LengthPrior* prior ) {
        std::string ruleText = _matchSmoothingRule( "LengthPrior", "*", "*" );
        double exponent = 0;

        int nextComma = 0;
        int nextColon = 0;
        int location = 0;

        for( location = 0; location < ruleText.length(); ) {
          nextComma = ruleText.find( ',', location );
          nextColon = ruleText.find( ':', location );

          std::string key = ruleText.substr( location, nextColon-location );
          std::string value = ruleText.substr( nextColon+1, nextComma-nextColon-1 );

          if( key == "exponent" ) {
            exponent = atof( value.c_str() );
          }
          if( nextComma > 0 )
            location = nextComma+1;
          else
            location = ruleText.size();
        }

        prior->setExponent( exponent );
      }
    };
  }
}

#endif // VANK_SMOOTHINGANNOTATORWALKER_HPP
