/*==========================================================================
 * Copyright (c) 2004 University of Massachusetts.  All Rights Reserved.
 *
 * Use of the Lemur Toolkit for Language Modeling and Information Retrieval
 * is subject to the terms of the software license set forth in the LICENSE
 * file included with this software, and also available at
 * http://www.lemurproject.org/license.html
 *
 *==========================================================================
 */

#include "vank/CachedFrequencyBeliefNode.hpp"

vank::infnet::CachedFrequencyBeliefNode::CachedFrequencyBeliefNode( const std::string& name,
                                                                     vank::lang::ListCache::CachedList* list,
                                                                     vank::query::TermScoreFunction& scoreFunction,
                                                                     double maximumBackgroundScore,
                                                                     double maximumScore )
  :
  _name(name),
  _list(list),
  _function(scoreFunction),
  _maximumBackgroundScore(maximumBackgroundScore),
  _maximumScore(maximumScore)
{
  _iter = _list->entries.begin();
}

lemur::api::DOCID_T vank::infnet::CachedFrequencyBeliefNode::nextCandidateDocument() {
  return _iter < _list->entries.end() ? _iter->document : MAX_INT32;
}

double vank::infnet::CachedFrequencyBeliefNode::maximumBackgroundScore() {
  return _maximumBackgroundScore;
}

double vank::infnet::CachedFrequencyBeliefNode::maximumScore() {
  return _maximumScore;
}

const vank::utility::greedy_vector<vank::api::ScoredExtentResult>& vank::infnet::CachedFrequencyBeliefNode::score( lemur::api::DOCID_T documentID, int begin, int end, int documentLength ) {
  assert( begin == 0 && end == documentLength ); // FrequencyListCopier ensures this condition
  const vank::index::DocumentContextCount* entry = _iter < _list->entries.end() ? _iter : 0;
  _extents.clear();

  int count = 0;
  int contextSize = 0;

  if( entry && entry->document == documentID ) {
    count = entry->count;
    contextSize = entry->contextSize;

    // advance this pointer so we're looking at the next document
    _iter++;
  } else {
    count = 0;
    contextSize = documentLength;
  }

  double score = _function.scoreOccurrence( count, contextSize );

  assert( score <= _maximumScore );
  _extents.push_back( vank::api::ScoredExtentResult( score, documentID, begin, end ) );

  return _extents;
}

bool vank::infnet::CachedFrequencyBeliefNode::hasMatch( lemur::api::DOCID_T documentID ) {
  return ( _iter < _list->entries.end() && _iter->document == documentID );
}

const vank::utility::greedy_vector<bool>& vank::infnet::CachedFrequencyBeliefNode::hasMatch( lemur::api::DOCID_T documentID, const vank::utility::greedy_vector<vank::index::Extent>& extents ) {
  // bogus result
  _matches.resize( extents.size(), false );
  return _matches;
}

const std::string& vank::infnet::CachedFrequencyBeliefNode::getName() const {
  return _name;
}

void vank::infnet::CachedFrequencyBeliefNode::annotate( vank::infnet::Annotator& annotator, lemur::api::DOCID_T documentID, int begin, int end ) {
  // can't annotate -- don't have position info
}

void vank::infnet::CachedFrequencyBeliefNode::indexChanged( vank::index::Index& index ) {
  // do nothing
}
