/*==========================================================================
 * Copyright (c) 2006 Carnegie Mellon University.  All Rights Reserved.
 *
 * Use of the Lemur Toolkit for Language Modeling and Information Retrieval
 * is subject to the terms of the software license set forth in the LICENSE
 * file included with this software (and below), and also available at
 * http://www.lemurproject.org/license.html
 *
 *==========================================================================
*/


//
// ExtentDescendantNode
//
// 31 Jan 2006 -- pto
//

#include "vank/ExtentDescendantNode.hpp"
#include "lemur/lemur-compat.hpp"
#include "vank/Annotator.hpp"

vank::infnet::ExtentDescendantNode::ExtentDescendantNode( const std::string& name,
                                                           ListIteratorNode* inner,
                                                           ListIteratorNode* outer,
                                                           DocumentStructureHolderNode & documentStructureHolderNode ) :
  _inner(inner),
  _outer(outer),
  _docStructHolder(documentStructureHolderNode),
  _name(name)
{
}

void vank::infnet::ExtentDescendantNode::prepare( lemur::api::DOCID_T documentID ) {
  // initialize the child / sibling pointer
  initpointer();

  _extents.clear();
  _lastExtent.begin = -1;
  _lastExtent.end = -1;

  if( !_inner || !_outer )
    return;

  const vank::utility::greedy_vector<vank::index::Extent>& inExtents = _inner->extents();
  const vank::utility::greedy_vector<vank::index::Extent>& outExtents = _outer->extents();

  vank::utility::greedy_vector<vank::index::Extent>::const_iterator innerIter = inExtents.begin();
  vank::utility::greedy_vector<vank::index::Extent>::const_iterator outerIter = outExtents.begin();

  vank::index::DocumentStructure * docStruct = _docStructHolder.getDocumentStructure();


  // check the inner extents, searching for an ancestor in outerNodes
  while ( innerIter != inExtents.end() ) {

    _leafs.clear();
    if ( innerIter->ordinal == 0 ) {
      docStruct->findLeafs( &_leafs, innerIter->begin, innerIter->end, true );
    } else {
      _leafs.insert( innerIter->ordinal );
    }

    std::set<int>::iterator leaf = _leafs.begin();
    bool found = false;
    while ( leaf != _leafs.end() && !found) {

      outerIter = outExtents.begin();
      while ( outerIter != outExtents.end() && !found ) {

        _ancestors.clear();
        if ( outerIter->ordinal == 0 ) {
          docStruct->findLeafs( &_ancestors, outerIter->begin, outerIter->end, true );
        } else {
          _ancestors.insert( outerIter->ordinal );
        }

        std::set<int>::iterator ancestor = _ancestors.begin();
        while ( ancestor != _ancestors.end() && !found ) {
          if ( docStruct->ancestor( *leaf, *ancestor ) ) {
            found = true;
            vank::index::Extent extent( innerIter->weight * outerIter->weight,
                                         innerIter->begin,
                                         innerIter->end,
                                         innerIter->ordinal );
            _extents.push_back( extent );
          }
          ancestor++;
        }
        outerIter++;
      }
      leaf++;
    }
    innerIter++;
  }

}

const vank::utility::greedy_vector<vank::index::Extent>& vank::infnet::ExtentDescendantNode::extents() {
  return _extents;
}

lemur::api::DOCID_T vank::infnet::ExtentDescendantNode::nextCandidateDocument() {
  return lemur_compat::max( _inner->nextCandidateDocument(), _outer->nextCandidateDocument() );
}

const std::string& vank::infnet::ExtentDescendantNode::getName() const {
  return _name;
}

void vank::infnet::ExtentDescendantNode::annotate( class Annotator& annotator, lemur::api::DOCID_T documentID, vank::index::Extent &extent ) {
  if (! _lastExtent.contains(extent)) {
    // if the last extent we annotated contains this one, there is no work
    // to do.
    _lastExtent = extent;

    annotator.addMatches( _extents, this, documentID, extent );

    vank::index::Extent range( extent.begin, extent.end );
    vank::utility::greedy_vector<vank::index::Extent>::const_iterator iter;
    iter = std::lower_bound( _extents.begin(), _extents.end(), range, vank::index::Extent::begins_before_less() );

    for( size_t i = iter-_extents.begin(); i<_extents.size() && _extents[i].begin <= extent.end; i++ ) {
      _inner->annotate( annotator, documentID, (vank::index::Extent &)_extents[i] );
      _outer->annotate( annotator, documentID, (vank::index::Extent &)_extents[i] );
    }
  }
}

void vank::infnet::ExtentDescendantNode::indexChanged( vank::index::Index& index ) {
  _lastExtent.begin = -1;
  _lastExtent.end = -1;
}

const vank::utility::greedy_vector<vank::index::Extent>& vank::infnet::ExtentDescendantNode::matches( vank::index::Extent &extent ) {

  _matches.clear();

  int begin = extent.begin;
  int end = extent.end;

  vank::index::DocumentStructure * docStruct = _docStructHolder.getDocumentStructure();

  _ancestors.clear();
  if ( extent.ordinal == 0 ) {
    docStruct->findLeafs( &_ancestors, begin, end, true);
  } else {
    _ancestors.insert( extent.ordinal );
  }

  for( size_t i = 0; i < _extents.size(); i++ ) {
    int innerBegin = _extents[i].begin;
    int innerEnd = _extents[i].end;

    _leafs.clear();
    if ( _extents[i].ordinal == 0 ) {
      docStruct->findLeafs( &_leafs, innerBegin, innerEnd, true);
    } else {
      _leafs.insert( _extents[i].ordinal );
    }

    bool match = false;
    std::set<int>::iterator ancestor = _ancestors.begin();
    while ( !match && ancestor != _ancestors.end() ) {
      std::set<int>::iterator leaf = _leafs.begin();
      while ( !match && leaf != _leafs.end() ) {
        if ( *leaf != *ancestor && docStruct->ancestor(*leaf, *ancestor) ) {
          match = true;
        }
        leaf++;
      }
      ancestor++;
    }
    if ( match ) {
      _matches.push_back(_extents[i]);
    }
  }

  return _matches;
}
