/*==========================================================================
 * Copyright (c) 2004 University of Massachusetts.  All Rights Reserved.
 *
 * Use of the Lemur Toolkit for Language Modeling and Information Retrieval
 * is subject to the terms of the software license set forth in the LICENSE
 * file included with this software, and also available at
 * http://www.lemurproject.org/license.html
 *
 *==========================================================================
 */

//
// FixedPassageNode
//
// 23 February 2005 -- tds
//

#include "vank/FixedPassageNode.hpp"
#include "vank/Annotator.hpp"
#include "lemur/lemur-compat.hpp"

vank::infnet::FixedPassageNode::FixedPassageNode( const std::string& name, vank::infnet::BeliefNode* child, int windowSize, int increment ) :
  _name(name),
  _child(child),
  _windowSize(windowSize),
  _increment(increment)
{
}

lemur::api::DOCID_T vank::infnet::FixedPassageNode::nextCandidateDocument() {
  return _child->nextCandidateDocument();
}

double vank::infnet::FixedPassageNode::maximumBackgroundScore() {
  return VANK_TINY_SCORE;
}

double vank::infnet::FixedPassageNode::maximumScore() {
  return VANK_HUGE_SCORE;
}

//
// score
//

const vank::utility::greedy_vector<vank::api::ScoredExtentResult>& vank::infnet::FixedPassageNode::score( lemur::api::DOCID_T documentID, vank::index::Extent &extent, int documentLength ) {
  // we're going to run through the field list, etc.
  _scores.clear();

  // find out how to chop up [begin,end] appropriately into passages
  _buildSubextents( extent );
  const vank::utility::greedy_vector<bool>& matches = _child->hasMatch( documentID, _subextents );

  // loop through the subextents, scoring each one
  for( size_t i=0; i<_subextents.size(); i++ ) {
    int scoreBegin = _subextents[i].begin;
    int scoreEnd = _subextents[i].end;
    if( !matches[i] )
      continue;

    const vank::utility::greedy_vector<vank::api::ScoredExtentResult>& childResults = _child->score( documentID, _subextents[i], documentLength );

    for( size_t j=0; j<childResults.size(); j++ ) {

      vank::api::ScoredExtentResult result(childResults[j]);
      result.document=documentID;
      result.begin=scoreBegin;
      result.end=scoreEnd;
      _scores.push_back( result );
    }
  }

  return _scores;
}

//
// annotate
//

void vank::infnet::FixedPassageNode::annotate( vank::infnet::Annotator& annotator, lemur::api::DOCID_T documentID, vank::index::Extent &extent ) {
  annotator.add(this, documentID, extent);

  // round down to find where the passage starts
  _buildSubextents( extent );
  const vank::utility::greedy_vector<bool>& matches = _child->hasMatch( documentID, _subextents );

  for( size_t i=0; i<_subextents.size(); i++ ) {
    if (matches[i])
      _child->annotate( annotator, documentID, _subextents[i] );
  }
}

//
// hasMatch
//

bool vank::infnet::FixedPassageNode::hasMatch( lemur::api::DOCID_T documentID ) {
  return _child->hasMatch( documentID );
}

//
// _addSubextents
//

void vank::infnet::FixedPassageNode::_addSubextents( const vank::index::Extent& extent ) {
  int beginPassage = (extent.begin / _increment) * _increment;
  int endPassage = beginPassage + _windowSize;

  while( beginPassage <= extent.end - _windowSize ) {
    int begin = lemur_compat::max( beginPassage, extent.begin );
    int end = lemur_compat::min( endPassage, extent.end );

    _subextents.push_back( vank::index::Extent( begin, end ) );

    beginPassage += _increment;
    endPassage = beginPassage + _windowSize;
  }

  // final passage; may overlap more than other passages
  if( beginPassage < extent.end ) {
    int begin = lemur_compat::max( extent.end - _windowSize, beginPassage );
    int end = extent.end;

    _subextents.push_back( vank::index::Extent( begin, end ) );
  }
}

//
// _buildSubextents
//

void vank::infnet::FixedPassageNode::_buildSubextents( const vank::index::Extent& extent ) {
  _subextents.clear();
  _addSubextents( extent );
}

//
// _buildSubextents
//

void vank::infnet::FixedPassageNode::_buildSubextents( const vank::utility::greedy_vector<vank::index::Extent>& extents ) {
  _subextents.clear();

  for( size_t i=0; i<extents.size(); i++ ) {
    _addSubextents( extents[i] );
  }
}

//
// hasMatch
//

const vank::utility::greedy_vector<bool>& vank::infnet::FixedPassageNode::hasMatch( lemur::api::DOCID_T documentID, const vank::utility::greedy_vector<vank::index::Extent>& extents ) {
  _matches.clear();
  _matches.resize( extents.size(), false );

  // to match, we split up each extent into its passage components, then check to see if there are sub-matches there.
  size_t i=0;
  size_t j=0;

  _buildSubextents( extents );

  // now that we have subextents, ask the child for regions that have results
  const vank::utility::greedy_vector<bool>& childMatches = _child->hasMatch( documentID, _subextents );

  // walk the extents and subextents simultaneously
  i = 0;
  j = 0;

  while( i < childMatches.size() && j < extents.size() ) {
    if( _subextents[i].begin < extents[j].begin ) {
      i++;
      continue;
    }

    if( _subextents[i].end > extents[j].end ) {
      j++;
      continue;
    }

    assert( _subextents[i].begin >= extents[j].begin );
    assert( _subextents[i].end <= extents[j].end );

    if( childMatches[i] ) {
      _matches[j] = true;
      // one match is sufficient, advance extents
      j++;
    } else {
      // it will never match, advance _subextents
      i++;
    }
  }

  return _matches;
}

const std::string& vank::infnet::FixedPassageNode::getName() const {
  return _name;
}

void vank::infnet::FixedPassageNode::indexChanged( vank::index::Index& index ) {
  // do nothing
}
