/*==========================================================================
 * Copyright (c) 2004 University of Massachusetts.  All Rights Reserved.
 *
 * Use of the Lemur Toolkit for Language Modeling and Information Retrieval
 * is subject to the terms of the software license set forth in the LICENSE
 * file included with this software, and also available at
 * http://www.lemurproject.org/license.html
 *
 *==========================================================================
 */

//
// IndexEnvironment
//
// 19 July 2004 -- tds
//

#include "vank/IndexEnvironment.hpp"
#include "vank/CompressedCollection.hpp"
#include "lemur/Exception.hpp"
#include "vank/AnchorTextAnnotator.hpp"
#include "vank/TaggedDocumentIterator.hpp"
#include "vank/VankParser.hpp"
#include "vank/VankTokenizer.hpp"
#include "vank/Path.hpp"
#include "vank/Conflater.hpp"
#include <iostream>

void vank::api::IndexEnvironment::_getParsingContext( vank::parse::Parser** parser,
                                                       vank::parse::Tokenizer** tokenizer,
                                                       vank::parse::DocumentIterator** iterDoc,
                                                       vank::parse::Conflater** conflater,
                                                       const std::string& className ) {
  std::string parserName;
  std::string iteratorName;

  *parser = 0;
  *tokenizer = 0;
  *iterDoc = 0;
  *conflater = 0;

  // look for an already-built environment
  std::map<std::string, vank::parse::FileClassEnvironment*>::iterator iter;
  iter = _environments.find(className);

  if( iter != _environments.end() ) {
    *parser = iter->second->parser;
    *tokenizer = iter->second->tokenizer;
    *iterDoc = iter->second->iterator;
    *conflater = iter->second->conflater;
    return;
  }

  vank::parse::FileClassEnvironment* fce = _fileClassFactory.get( className );

  if( fce ) {
    _environments[className] = fce;
    *parser = fce->parser;
    *tokenizer = fce->tokenizer;
    *iterDoc = fce->iterator;
    *conflater = fce->conflater;
  }
}

std::vector<vank::parse::Transformation*> vank::api::IndexEnvironment::_createAnnotators( const std::string& fileName,
                                                                                            const std::string& fileClass,
                                                                                            vank::parse::Conflater** conflater ) {
  vank::parse::AnchorTextAnnotator* annotator = 0;
  vank::parse::OffsetAnnotationAnnotator* oa_annotator = 0;
  vank::parse::OffsetMetadataAnnotator* om_annotator = 0;

  std::vector<vank::parse::Transformation*> annotators;
  std::string relativePath;

  // get the path to the filename relative to the document root
  if (fileName.empty()) {
    // case: addString
    relativePath = "";
  } else {
    // case: addFile
    relativePath = vank::file::Path::relative( _documentRoot, fileName );
  }

  if( _anchorTextRoot.length() ) {
    // if the user specified some anchor text, we'll add it in
    std::string anchorTextPath;
    if( relativePath.length() > 0 ) {
      anchorTextPath = vank::file::Path::combine( _anchorTextRoot, relativePath );
    } else {
      anchorTextPath = _anchorTextRoot;
    }
    _annotator.open( anchorTextPath );
    annotator = &_annotator;
    annotators.push_back( annotator );
  }

  if( _offsetAnnotationsRoot.length() ) {
    // If the user specified an offset annotations file, we'll use it.
    std::string offsetAnnotationsPath;
    // only do path combining on directories. Otherwise, assume a single file
    // for all corpus documents.
    if( vank::file::Path::isDirectory( _offsetAnnotationsRoot ) && relativePath.length() > 0 ) {
      offsetAnnotationsPath = vank::file::Path::combine( _offsetAnnotationsRoot, relativePath );
    } else {
      offsetAnnotationsPath = _offsetAnnotationsRoot;
    }
    _oa_annotator.setConflater( *conflater );
    _oa_annotator.open( offsetAnnotationsPath );
    oa_annotator = &_oa_annotator;
    annotators.push_back( oa_annotator );
  }

  if( _offsetMetadataRoot.length() ) {
    // If the user specified an offset metadata path, we'll use it.
    std::string offsetMetadataPath;
    // only do path combining on directories. Otherwise, assume a single file
    // for all corpus documents.
    if( vank::file::Path::isDirectory( _offsetMetadataRoot ) && relativePath.length() > 0 ) {
      offsetMetadataPath = vank::file::Path::combine( _offsetMetadataRoot, relativePath );
    } else {
      offsetMetadataPath = _offsetMetadataRoot;
    }
    _om_annotator.open( offsetMetadataPath );
    om_annotator = &_om_annotator;
    annotators.push_back( om_annotator );
  }

  return annotators;
}

vank::api::ParsedDocument* vank::api::IndexEnvironment::_applyAnnotators( std::vector<vank::parse::Transformation*>& annotators,
                                                                            vank::api::ParsedDocument* parsed ) {
  for( size_t i=0; i<annotators.size(); i++ ) {
    if( annotators[i] ) {
      parsed = annotators[i]->transform( parsed );
    }
  }

  return parsed;
}

vank::api::IndexEnvironment::IndexEnvironment() :
  _callback(0),
  _options(0),
  _maxdocs(1 << 8),
  _documentsIndexed(0),
  _documentsSeen(0)
{
}

vank::api::IndexEnvironment::~IndexEnvironment() {
  close();
  vank::utility::delete_map_contents<std::string, vank::parse::FileClassEnvironment>( _environments );
}

int vank::api::IndexEnvironment::setBlackList( const std::string& blackListFile)
{
    ifstream inf(blackListFile.data(), ifstream::in);

    std::string line;

    while(getline(inf, line))
    {
        _blackedDocs.insert(line);
    }

    inf.close();

    return _blackedDocs.size();
}


void vank::api::IndexEnvironment::setNormalization( bool flag ) {
  _parameters.set( "normalize", flag );
}

void vank::api::IndexEnvironment::setStoreDocs( bool flag ) {
  _parameters.set( "storeDocs", flag );
}

void vank::api::IndexEnvironment::setInjectURL( bool flag ) {
  _parameters.set( "injectURL", flag );
}

void vank::api::IndexEnvironment::setMemory( UINT64 memory ) {
  _parameters.set("memory", memory);
}

void vank::api::IndexEnvironment::setMaxDocs( UINT64 maxdocs ) {
  _maxdocs = maxdocs;
}

bool vank::api::IndexEnvironment::haveReachedDocLimit() {
  return (_maxdocs <= _documentsIndexed);
}

void vank::api::IndexEnvironment::setOffsetAnnotationsPath( const std::string& offsetAnnotationsRoot ) {
  _offsetAnnotationsRoot = offsetAnnotationsRoot;
}

void vank::api::IndexEnvironment::setOffsetMetadataPath( const std::string& offsetMetadataRoot ) {
  _offsetMetadataRoot = offsetMetadataRoot;
}

void vank::api::IndexEnvironment::setAnchorTextPath( const std::string& anchorTextRoot ) {
  _anchorTextRoot = anchorTextRoot;
}

void vank::api::IndexEnvironment::setDocumentRoot( const std::string& documentRoot ) {
  _documentRoot = documentRoot;
}

void vank::api::IndexEnvironment::setStopwords( const std::vector<std::string>& stopwords ) {
  _parameters.set("stopper","");
  Parameters p = _parameters.get("stopper");
  for( unsigned int i=0; i<stopwords.size(); i++ ) {
    p.append("word").set(stopwords[i]);
  }
}

void vank::api::IndexEnvironment::setStemmer( const std::string& stemmer ) {
  _parameters.set("stemmer.name", stemmer);
}

void vank::api::IndexEnvironment::setOffsetAnnotationIndexHint(vank::parse::OffsetAnnotationIndexHint hintType) {
  _oa_annotator.setHint(hintType);
}

void vank::api::IndexEnvironment::addFileClass( const std::string& name,
                                                 const std::string& iter,
                                                 const std::string& parser,
                                                 const std::string& tokenizer,
                                                 const std::string& startDocTag,
                                                 const std::string& endDocTag,
                                                 const std::string& endMetadataTag,
                                                 const std::vector<std::string>& include,
                                                 const std::vector<std::string>& exclude,
                                                 const std::vector<std::string>& index,
                                                 const std::vector<std::string>& metadata,
                                                 const std::map<vank::parse::ConflationPattern*,std::string>& conflations )
{
  this->_fileClassFactory.addFileClass( name, iter, parser, tokenizer, startDocTag, endDocTag, endMetadataTag,
                                        include, exclude, index, metadata, conflations );

}

void vank::api::IndexEnvironment::setIndexedFields( const std::vector<std::string>& fieldNames ) {
  bool existingFields = _parameters.exists( "field" );
  // if setNumericField or setOrdinalField was called before setIndexedFields, there
  // will be an existing entry in the fields.
  for( unsigned int i=0; i<fieldNames.size(); i++) {
    bool found = false;
    if( existingFields ) {
      Parameters fields = _parameters["field"];
      for( size_t j=0; j<fields.size(); j++ ) {
        std::string parameterFieldName = fields[j]["name"];
        if( parameterFieldName == fieldNames[i] ) {
          // already there, don't add
          found = true;
        }
      }
    }
    if (! found)
      _parameters.append("field").set("name",fieldNames[i]);
  }
}

void vank::api::IndexEnvironment::setNumericField( const std::string& fieldName, bool isNumeric, const std::string & parserName ) {
  bool existingFields = _parameters.exists( "field" );

  if (existingFields) {
    Parameters fields = _parameters["field"];
    for( size_t i=0; i<fields.size(); i++ ) {
      std::string parameterFieldName = fields[i]["name"];

      if( parameterFieldName == fieldName ) {
        fields[i].set( "numeric", isNumeric );
        fields[i].set( "parserName", parserName );
        return;
      }
    }
  }
  Parameters field = _parameters.append("field");
  field.set( "name", fieldName );
  field.set( "numeric", isNumeric );
  field.set( "parserName", parserName );
}

void vank::api::IndexEnvironment::setOrdinalField( const std::string& fieldName, bool isOrdinal ) {
  bool existingFields = _parameters.exists( "field" );

  if ( existingFields ) {
    Parameters fields = _parameters["field"];

    for( size_t i=0; i<fields.size(); i++ ) {
      std::string parameterFieldName = fields[i]["name"];

      if( parameterFieldName == fieldName ) {
        fields[i].set( "ordinal", isOrdinal );
        return;
      }
    }
  }
  Parameters field = _parameters.append("field");
  field.set( "name", fieldName );
  field.set( "ordinal", isOrdinal );
}

void vank::api::IndexEnvironment::setParentalField( const std::string& fieldName, bool isParental ) {
  bool existingFields = _parameters.exists( "field" );

  if ( existingFields ) {
    Parameters fields = _parameters["field"];

    for( int i=0; i<fields.size(); i++ ) {
      std::string parameterFieldName = fields[i]["name"];

      if( parameterFieldName == fieldName ) {
        fields[i].set( "parental", isParental );
        return;
      }
    }
  }
  Parameters field = _parameters.append("field");
  field.set( "name", fieldName );
  field.set( "parental", isParental );
}


//
// setMetadataIndexedFields
//

void vank::api::IndexEnvironment::setMetadataIndexedFields( const std::vector<std::string>& forwardFields, const std::vector<std::string>& backwardFields ) {
  if( !_parameters.exists("collection") )
    _parameters.set("collection", "");

  Parameters collection = _parameters.get("collection");
  for( unsigned int i=0; i<forwardFields.size(); i++ ) {
    if( i==0 )
      collection.set("forward", forwardFields[i]);
    else
      collection.append("forward").set(forwardFields[i]);
  }

  for( unsigned int i=0; i<backwardFields.size(); i++ ) {
    if( i==0 )
      collection.set("backward", backwardFields[i]);
    else
      collection.append("backward").set(backwardFields[i]);
  }
}

//
// create
//

void vank::api::IndexEnvironment::create( const std::string& repositoryPath, vank::api::IndexStatus* callback ) {
  _callback = callback;
  _repository.create( repositoryPath, &_parameters );
}

//
// open
//

void vank::api::IndexEnvironment::open( const std::string& repositoryPath, vank::api::IndexStatus* callback ) {
  _callback = callback;
  _repository.open( repositoryPath, &_parameters );
}

//
// close
//

void vank::api::IndexEnvironment::close() {
  _repository.close();
}

//
// addFile
//

void vank::api::IndexEnvironment::addFile( const std::string& fileName, vank::api::IndexDocumentMetaCallback* metaCB ) {
  std::string extension = vank::file::Path::extension( fileName );
  addFile( fileName, extension, metaCB );
}

//
// addFile
//

void vank::api::IndexEnvironment::addFile( const std::string& fileName, const std::string& fileClass, vank::api::IndexDocumentMetaCallback* metaCB ) {
  vank::parse::Parser* parser = 0;
  vank::parse::Tokenizer* tokenizer = 0;
  vank::parse::DocumentIterator* iterator = 0;
  vank::parse::Conflater* conflater = 0;

  if( haveReachedDocLimit() ) {
    //documentID = -2; addFile return is void
    return;
  }

  _getParsingContext( &parser, &tokenizer, &iterator, &conflater, fileClass );

  if( !parser || !iterator ) {
    _documentsSeen++;
    if( _callback ) (*_callback) ( vank::api::IndexStatus::FileSkip, fileName, _error, _documentsIndexed, _documentsSeen );
  } else {
    try {
      vank::parse::UnparsedDocument* document;
      vank::parse::TokenizedDocument* tokenized;
      ParsedDocument* parsed;

      iterator->open( fileName );

      std::vector<vank::parse::Transformation*> annotators = _createAnnotators( fileName, fileClass, &conflater );

      // notify caller that the file was successfully parsed
      if( _callback ) (*_callback)( vank::api::IndexStatus::FileOpen, fileName, _error, _documentsIndexed, _documentsSeen );

      while( (document = iterator->nextDocument()) ) {
        _documentsSeen++;

        tokenized = tokenizer->tokenize( document );
        parsed = parser->parse( tokenized );
        /*
         * can't know for sure that we have the docno element until after the annotators have been applied
         * check if this document is in the index already, duplicates must not be indexed
         */
        parsed = _applyAnnotators( annotators, parsed );
        // find the docno
        std::string docIDStr = "";
        for( size_t i=0; i<parsed->metadata.size(); i++ ) {
          const char * key = parsed->metadata[i].key;
          if( !strcmp( key, "docno" ) ) docIDStr = (const char *)parsed->metadata[i].value;
        }

        // Check if blacklisted
        if(_blackedDocs.find(docIDStr)!=_blackedDocs.end())
            continue;

        // look up the id.
        std::vector<lemur::api::DOCID_T> ids = _repository.collection()->retrieveIDByMetadatum("docno", docIDStr);
        // if not found, add the document.
        if (ids.size() == 0)  {
            // allow caller to adjust document metadata before we add document
            // if( metaCB ) (*metaCB)( vank::api::IndexDocumentMetaCallback::FileSource, fileClass, parsed->metadata );
            _repository.addDocument( parsed );
            _documentsIndexed++;
        } else {
          // duplicate, don't index, return 0 as doc id
          //documentID = -1; addFile return is void
        }

        if( _callback ) (*_callback)( vank::api::IndexStatus::DocumentCount, fileName, _error, _documentsIndexed, _documentsSeen );
      }

      // notify caller that the file was successfully closed
      if( _callback ) (*_callback)( vank::api::IndexStatus::FileClose, fileName, _error, _documentsIndexed, _documentsSeen );

      iterator->close();
    } catch( lemur::api::Exception& e ) {
      if( iterator )
        iterator->close();

      // notify caller of errors
      if( _callback ) (*_callback)( vank::api::IndexStatus::FileError, fileName, e.what(), _documentsIndexed, _documentsSeen );
    }
  }
}

//
// addString
//

lemur::api::DOCID_T vank::api::IndexEnvironment::addString( const std::string& documentString, const std::string& fileClass, const std::vector<vank::parse::MetadataPair>& metadata ) {
  vank::parse::UnparsedDocument document;
  vank::parse::Parser* parser = 0;
  vank::parse::Tokenizer* tokenizer = 0;
  vank::parse::DocumentIterator* iterator = 0;
  vank::parse::Conflater* conflater = 0;
  std::string nothing = "";
  lemur::api::DOCID_T documentID = 0; // invalid

  if( haveReachedDocLimit() ) {
    //documentID = -2; addFile return is void
    return -2;
  }

  std::vector<vank::parse::Transformation*> annotators = _createAnnotators( nothing, fileClass, &conflater );

  _documentsSeen++;

  document.text = documentString.c_str();
  document.textLength = documentString.length() + 1; // for the null
  document.metadata = metadata;
  document.content = document.text;
  document.contentLength = document.textLength - 1;

  _getParsingContext( &parser, &tokenizer, &iterator, &conflater, fileClass );

  if( parser == 0 ) {
    LEMUR_THROW( LEMUR_RUNTIME_ERROR, "File class '" + fileClass + "' wasn't recognized." );
  }
  vank::parse::TokenizedDocument* tokenized = tokenizer->tokenize( &document );

  ParsedDocument* parsed = parser->parse( tokenized );

  /*
   * can't know for sure that we have the docno element until after the annotators have been applied
   * check if this document is in the index already, duplicates must not be indexed
   */
  parsed = _applyAnnotators( annotators, parsed );
  // find the docno
  std::string docIDStr = "";
  for( size_t i=0; i<parsed->metadata.size(); i++ ) {
    const char * key = parsed->metadata[i].key;
    if( !strcmp( key, "docno" ) ) docIDStr = (const char *)parsed->metadata[i].value;
  }
  // look up the doc id
  std::vector<lemur::api::DOCID_T> ids = _repository.collection()->retrieveIDByMetadatum("docno", docIDStr);
  // if not found, add the document
  if (ids.size() == 0)  {
    documentID =_repository.addDocument( parsed );
    _documentsIndexed++;
  } else {
    // duplicete, don't index, return 0 as doc id
    documentID = -1;
  }

  if( _callback ) (*_callback)( vank::api::IndexStatus::DocumentCount, nothing, _error, _documentsIndexed, _documentsSeen );

  return documentID;
}

//
// For UIMA with offset annotations
//
lemur::api::DOCID_T vank::api::IndexEnvironment::addString( const std::string& documentString, const std::string&
                                             fileClass, const std::vector<vank::parse::MetadataPair>& metadata, const std::vector<vank::parse::TagExtent *> &tags ) {
  vank::parse::UnparsedDocument document;
  vank::parse::Parser* parser;
  vank::parse::Tokenizer* tokenizer;
  vank::parse::DocumentIterator* iterator;
  vank::parse::Conflater* conflater;
  vank::parse::OffsetAnnotationAnnotator *annote;
  std::string docno = "";

  for ( size_t i=0; i<metadata.size(); i++ ) {
    const char* attributeName = metadata[i].key;
    const char* attributeValue = (const char*) metadata[i].value;
    if ( ! strcmp( attributeName, "docno" ) ) docno = attributeValue;
  }

  std::string nothing;

  _documentsSeen++;

  document.text = documentString.c_str();
  document.textLength = documentString.length() + 1; // for the null
  document.metadata = metadata;
  document.content = document.text;
  document.contentLength = document.textLength - 1;

  _getParsingContext( &parser, &tokenizer, &iterator, &conflater, fileClass );
  annote = new vank::parse::OffsetAnnotationAnnotator(conflater);
  annote->setTags(docno.c_str(), tags);

  if( parser == 0 ) {
    LEMUR_THROW( LEMUR_RUNTIME_ERROR, "File class '" + fileClass + "' wasn't recognized." );
  }
  vank::parse::TokenizedDocument* tokenized = tokenizer->tokenize( &document );

  ParsedDocument* parsed = parser->parse( tokenized );
  parsed = annote->transform(parsed);

  lemur::api::DOCID_T documentID =_repository.addDocument( parsed );

  _documentsIndexed++;
  if( _callback ) (*_callback)( vank::api::IndexStatus::DocumentCount, nothing, _error, _documentsIndexed, _documentsSeen );

  return documentID;
}


//
// addParsedDocument
//

lemur::api::DOCID_T vank::api::IndexEnvironment::addParsedDocument( ParsedDocument* document ) {
  std::string nothing;

  _documentsSeen++;
  lemur::api::DOCID_T documentID = _repository.addDocument( document );
  _documentsIndexed++;
  if( _callback ) (*_callback)( vank::api::IndexStatus::DocumentCount, nothing, _error, _documentsIndexed, _documentsSeen );

  return documentID;
}

//
// deleteDocument
//

void vank::api::IndexEnvironment::deleteDocument( lemur::api::DOCID_T documentID ) {
  _repository.deleteDocument( documentID );
}

//
// documentExists
//

bool vank::api::IndexEnvironment::documentExists( lemur::api::DOCID_T documentID ) {
  return _repository.documentExists( documentID );
}

//
// documentsIndexed
//

int vank::api::IndexEnvironment::documentsIndexed() {
  return _documentsIndexed;
}

//
// documentsSeen
//

int vank::api::IndexEnvironment::documentsSeen() {
  return _documentsSeen;
}

//
// compact
//

void vank::api::IndexEnvironment::compact() {
  _repository.compact();
}

//
// merge
//

void vank::api::IndexEnvironment::merge( const std::string& outputIndex, const std::vector<std::string>& inputIndexes ) {
  vank::collection::Repository::merge( outputIndex, inputIndexes );
}
