/*==========================================================================
 * Copyright (c) 2004 University of Massachusetts.  All Rights Reserved.
 *
 * Use of the Lemur Toolkit for Language Modeling and Information Retrieval
 * is subject to the terms of the software license set forth in the LICENSE
 * file included with this software, and also available at
 * http://www.lemurproject.org/license.html
 *
 *==========================================================================
 */


//
// LocalQueryServer
//
// 15 March 2004 -- tds
//

#include "vank/LocalQueryServer.hpp"
#include "vank/QuerySpec.hpp"
#include "lemur/lemur-platform.h"
#include "lemur/lemur-compat.hpp"
#include <vector>

#include "vank/UnnecessaryNodeRemoverCopier.hpp"
#include "vank/ContextSimpleCountCollectorCopier.hpp"
#include "vank/FrequencyListCopier.hpp"
#include "vank/DagCopier.hpp"

#include "vank/InferenceNetworkBuilder.hpp"
#include "vank/InferenceNetwork.hpp"

#include "vank/CompressedCollection.hpp"
#include "vank/delete_range.hpp"
#include "vank/WeightFoldingCopier.hpp"

#include "vank/Appliers.hpp"
#include "vank/ScopedLock.hpp"

#include "vank/TreePrinterWalker.hpp"

#include "vank/DocumentStructure.hpp"

//
// Response objects
//
namespace vank
{
  namespace server
  {

    class LocalQueryServerResponse : public QueryServerResponse {
    private:
      vank::infnet::InferenceNetwork::MAllResults _results;

    public:
      LocalQueryServerResponse( const vank::infnet::InferenceNetwork::MAllResults& results ) :
        _results(results) {
      }

      vank::infnet::InferenceNetwork::MAllResults& getResults() {
        return _results;
      }
    };

    class LocalQueryServerDocumentsResponse : public QueryServerDocumentsResponse {
    private:
      std::vector<vank::api::ParsedDocument*> _documents;

    public:
      LocalQueryServerDocumentsResponse( const std::vector<vank::api::ParsedDocument*>& results )
        :
        _documents(results)
      {
      }

      // caller's responsibility to delete these results
      std::vector<vank::api::ParsedDocument*>& getResults() {
        return _documents;
      }
    };

    class LocalQueryServerMetadataResponse : public QueryServerMetadataResponse {
    private:
      std::vector<std::string> _metadata;

    public:
      LocalQueryServerMetadataResponse( const std::vector<std::string>& metadata ) :
        _metadata(metadata)
      {
      }

      std::vector<std::string>& getResults() {
        return _metadata;
      }
    };

    class LocalQueryServerVectorsResponse : public QueryServerVectorsResponse {
    private:
      std::vector<vank::api::DocumentVector*> _vectors;

    public:
      LocalQueryServerVectorsResponse( int vectorCount ) {
        _vectors.reserve( vectorCount );
      }

      void addVector( vank::api::DocumentVector* vec ) {
        _vectors.push_back( vec );
      }

      // caller deletes vank::api::DocumentVector objects
      std::vector<vank::api::DocumentVector*>& getResults() {
        return _vectors;
      }
    };

    class LocalQueryServerDocumentIDsResponse : public QueryServerDocumentIDsResponse {
    private:
      std::vector<lemur::api::DOCID_T> _documentIDs;

    public:
      LocalQueryServerDocumentIDsResponse( const std::vector<lemur::api::DOCID_T>& documents ) :
        _documentIDs(documents)
      {
      }

      std::vector<lemur::api::DOCID_T>& getResults() {
        return _documentIDs;
      }
    };
  }
}

//
// Class code
//

vank::server::LocalQueryServer::LocalQueryServer( vank::collection::Repository& repository ) :
  _repository(repository), _maxWildcardMatchesPerTerm(vank::infnet::InferenceNetworkBuilder::DEFAULT_MAX_WILDCARD_TERMS)
{
  // if supplied and false, turn off optimization for all queries.
  _optimizeParameter = vank::api::Parameters::instance().get( "optimize", true );
}

//
// _indexWithDocument
//

vank::index::Index* vank::server::LocalQueryServer::_indexWithDocument( vank::collection::Repository::index_state& indexes, lemur::api::DOCID_T documentID ) {
  for( size_t i=0; i<indexes->size(); i++ ) {
    vank::thread::ScopedLock lock( (*indexes)[i]->statisticsLock() );
    lemur::api::DOCID_T lowerBound = (*indexes)[i]->documentBase();
    lemur::api::DOCID_T upperBound = (*indexes)[i]->documentMaximum();

    if( lowerBound <= documentID && upperBound > documentID ) {
      return (*indexes)[i];
    }
  }

  return 0;
}

//
// document
//

vank::api::ParsedDocument* vank::server::LocalQueryServer::document( lemur::api::DOCID_T documentID ) {
  vank::collection::CompressedCollection* collection = _repository.collection();
  vank::api::ParsedDocument* document = collection->retrieve( documentID );
  return document;
}

std::string vank::server::LocalQueryServer::documentMetadatum( lemur::api::DOCID_T documentID, const std::string& attributeName ) {
  vank::collection::CompressedCollection* collection = _repository.collection();
  return collection->retrieveMetadatum( documentID, attributeName );
}

vank::server::QueryServerMetadataResponse* vank::server::LocalQueryServer::documentMetadata( const std::vector<lemur::api::DOCID_T>& documentIDs, const std::string& attributeName ) {
  std::vector<std::string> result;

  std::vector<std::pair<lemur::api::DOCID_T, int> > docSorted;
  for( size_t i=0; i<documentIDs.size(); i++ ) {
    docSorted.push_back( std::make_pair( documentIDs[i], i ) );
  }
  std::sort( docSorted.begin(), docSorted.end() );

  for( size_t i=0; i<docSorted.size(); i++ ) {
    result.push_back( documentMetadatum(docSorted[i].first, attributeName) );
  }

  std::vector<std::string> actual;
  actual.resize( documentIDs.size() );
  for( size_t i=0; i<docSorted.size(); i++ ) {
    actual[docSorted[i].second] = result[i];
  }

  return new vank::server::LocalQueryServerMetadataResponse( actual );
}

vank::server::QueryServerDocumentsResponse* vank::server::LocalQueryServer::documents( const std::vector<lemur::api::DOCID_T>& documentIDs ) {
  std::vector<vank::api::ParsedDocument*> result;
  for( size_t i=0; i<documentIDs.size(); i++ ) {
    result.push_back( document(documentIDs[i]) );
  }
  return new vank::server::LocalQueryServerDocumentsResponse( result );
}

vank::server::QueryServerDocumentsResponse* vank::server::LocalQueryServer::documentsFromMetadata( const std::string& attributeName, const std::vector<std::string>& attributeValues ) {
  vank::collection::CompressedCollection* collection = _repository.collection();
  std::vector<vank::api::ParsedDocument*> result;

  for( size_t i=0; i<attributeValues.size(); i++ ) {
    std::vector<vank::api::ParsedDocument*> documents = collection->retrieveByMetadatum( attributeName, attributeValues[i] );
    std::copy( documents.begin(), documents.end(), std::back_inserter( result ) );
  }

  return new vank::server::LocalQueryServerDocumentsResponse( result );
}

vank::server::QueryServerDocumentIDsResponse* vank::server::LocalQueryServer::documentIDsFromMetadata( const std::string& attributeName, const std::vector<std::string>& attributeValues ) {
  vank::collection::CompressedCollection* collection = _repository.collection();
  std::vector<lemur::api::DOCID_T> result;

  for( size_t i=0; i<attributeValues.size(); i++ ) {
    std::vector<lemur::api::DOCID_T> documents = collection->retrieveIDByMetadatum( attributeName, attributeValues[i] );
    std::copy( documents.begin(), documents.end(), std::back_inserter( result ) );
  }

  return new vank::server::LocalQueryServerDocumentIDsResponse( result );
}

INT64 vank::server::LocalQueryServer::termCount() {
  vank::collection::Repository::index_state indexes = _repository.indexes();
  INT64 total = 0;

  for( size_t i=0; i<indexes->size(); i++ ) {
    vank::thread::ScopedLock lock( (*indexes)[i]->statisticsLock() );
    total += (*indexes)[i]->termCount();
  }

  return total;
}

INT64 vank::server::LocalQueryServer::termCountUnique() {
  vank::collection::Repository::index_state indexes = _repository.indexes();
  INT64 total = 0;

  for( size_t i=0; i<indexes->size(); i++ ) {
    vank::thread::ScopedLock lock( (*indexes)[i]->statisticsLock() );
    total += (*indexes)[i]->uniqueTermCount();
  }

  return total;
}

INT64 vank::server::LocalQueryServer::termCount( const std::string& term ) {
  std::string stem = _repository.processTerm( term );
  // stopwords return a string of length 0, causing Keyfile to throw.
  if( stem.length() != 0 ) {
    return stemCount( stem );
  } else {
    return 0;
  }
}

INT64 vank::server::LocalQueryServer::stemCount( const std::string& stem ) {
  vank::collection::Repository::index_state indexes = _repository.indexes();
  INT64 total = 0;

  for( size_t i=0; i<indexes->size(); i++ ) {
    vank::thread::ScopedLock lock( (*indexes)[i]->statisticsLock() );
    total += (*indexes)[i]->termCount( stem );
  }

  return total;
}

INT64 vank::server::LocalQueryServer::termFieldCount( const std::string& term, const std::string& field ) {
  std::string stem = _repository.processTerm( term );

  if( stem.length() != 0 ) {
    return stemFieldCount( stem, field );
  } else {
    return 0;
  }
}

INT64 vank::server::LocalQueryServer::stemFieldCount( const std::string& stem, const std::string& field ) {
  vank::collection::Repository::index_state indexes = _repository.indexes();
  INT64 total = 0;

  for( size_t i=0; i<indexes->size(); i++ ) {
    vank::thread::ScopedLock lock( (*indexes)[i]->statisticsLock() );
    total += (*indexes)[i]->fieldTermCount( field, stem );
  }

  return total;
}

std::string vank::server::LocalQueryServer::termName( lemur::api::TERMID_T term ) {
  vank::collection::Repository::index_state indexes = _repository.indexes();
  vank::index::Index* index = (*indexes)[0];
  vank::thread::ScopedLock lock( index->statisticsLock() );
  return index->term( term );
}

lemur::api::TERMID_T vank::server::LocalQueryServer::termID( const std::string& term ) {
  vank::collection::Repository::index_state indexes = _repository.indexes();
  vank::index::Index* index = (*indexes)[0];
  std::string processed = _repository.processTerm( term );
  vank::thread::ScopedLock lock( index->statisticsLock() );

  if( processed.length() != 0 ) {
    return index->term( processed.c_str() );
  } else {
    return 0;
  }
}

std::string vank::server::LocalQueryServer::stemTerm( const std::string& term ) {
  std::string stem  = _repository.processTerm(term);
  return stem;
}

std::vector<std::string> vank::server::LocalQueryServer::fieldList() {
  std::vector<std::string> result;
  const std::vector<vank::collection::Repository::Field>& fields = _repository.fields();

  for( size_t i=0; i<fields.size(); i++ ) {
    result.push_back( fields[i].name );
  }

  return result;
}

int vank::server::LocalQueryServer::documentLength( lemur::api::DOCID_T documentID ) {
  vank::collection::Repository::index_state indexes = _repository.indexes();
  vank::index::Index* index = _indexWithDocument( indexes, documentID );

  if( index ) {
    vank::thread::ScopedLock lock( index->statisticsLock() );
    return index->documentLength( documentID );
  }

  return 0;
}

INT64 vank::server::LocalQueryServer::documentCount() {
  vank::collection::Repository::index_state indexes = _repository.indexes();
  INT64 total = 0;

  for( size_t i=0; i<indexes->size(); i++ ) {
    vank::thread::ScopedLock lock( (*indexes)[i]->statisticsLock() );
    total += (*indexes)[i]->documentCount();
  }

  total -= _repository.deletedList().deletedCount();
  return total;
}

INT64 vank::server::LocalQueryServer::documentCount( const std::string& term ) {
  std::string stem = _repository.processTerm( term );
  return documentStemCount(stem);
}

INT64 vank::server::LocalQueryServer::documentStemCount( const std::string& stem ) {
  vank::collection::Repository::index_state indexes = _repository.indexes();
  INT64 total = 0;
  if( stem.length() == 0 ) return total;
  for( size_t i=0; i<indexes->size(); i++ ) {
    vank::thread::ScopedLock lock( (*indexes)[i]->statisticsLock() );
    total += (*indexes)[i]->documentCount( stem );
  }
  return total;
}

vank::server::QueryServerResponse* vank::server::LocalQueryServer::runQuery( std::vector<vank::lang::Node*>& roots, int resultsRequested, bool optimize ) {

  vank::lang::TreePrinterWalker printer;

  // use UnnecessaryNodeRemover to get rid of window nodes, ExtentAnd nodes and ExtentOr nodes
  // that only have one child and LengthPrior nodes where the exponent is zero
  vank::lang::ApplyCopiers<vank::lang::UnnecessaryNodeRemoverCopier> unnecessary( roots );

  // run the contextsimplecountcollectorcopier to gather easy stats
  vank::lang::ApplyCopiers<vank::lang::ContextSimpleCountCollectorCopier> contexts( unnecessary.roots(), _repository );

  // use frequency-only nodes where appropriate
  vank::lang::ApplyCopiers<vank::lang::FrequencyListCopier> frequency( contexts.roots(), _cache );

  // fold together any nested weight nodes
  vank::lang::ApplyCopiers<vank::lang::WeightFoldingCopier> weight( frequency.roots() );

  // make all this into a dag
  vank::lang::ApplySingleCopier<vank::lang::DagCopier> dag( weight.roots(), _repository );

  std::vector<vank::lang::Node*>& networkRoots = dag.roots();
  // turn off optimization if called with optimize == false
  // turn off optimization if called the Parameter optimize == false
  if( !optimize || !_optimizeParameter ) {
    // we may be asked not to perform optimizations that might
    // drastically change the structure of the tree; for instance,
    // annotation queries may ask for this
    networkRoots = contexts.roots();
  }
  /*
    vank::lang::TreePrinterWalker printer;
    vank::lang::ApplyWalker<vank::lang::TreePrinterWalker> printTree(networkRoots, &printer);
  */

  // build an inference network
  vank::infnet::InferenceNetworkBuilder builder( _repository, _cache, resultsRequested, _maxWildcardMatchesPerTerm );
  vank::lang::ApplyWalker<vank::infnet::InferenceNetworkBuilder> buildWalker( networkRoots, &builder );

  vank::infnet::InferenceNetwork* network = builder.getNetwork();
  vank::infnet::InferenceNetwork::MAllResults result;
  result = network->evaluate();

  return new vank::server::LocalQueryServerResponse( result );
}

vank::server::QueryServerVectorsResponse* vank::server::LocalQueryServer::documentVectors( const std::vector<lemur::api::DOCID_T>& documentIDs ) {
  vank::server::LocalQueryServerVectorsResponse* response = new vank::server::LocalQueryServerVectorsResponse( (int)documentIDs.size() );
  vank::collection::Repository::index_state indexes = _repository.indexes();
  std::map<int, std::string> termIDStringMap;

  for( size_t i=0; i<documentIDs.size(); i++ ) {
    vank::index::Index* index = _indexWithDocument( indexes, documentIDs[i] );

    {
      vank::thread::ScopedLock lock( index->statisticsLock() );

      const vank::index::TermList* termList = index->termList( documentIDs[i] );
      vank::api::DocumentVector* result = new vank::api::DocumentVector( index, termList, termIDStringMap );
      delete termList;
      response->addVector( result );
    }
  }

  return response;
}

vank::server::QueryServerMetadataResponse* vank::server::LocalQueryServer::pathNames( const std::vector<lemur::api::DOCID_T>& documentIDs, const std::vector<int>& pathBegins, const std::vector<int>& pathEnds ) {

  int lastDoc = 0;
  vank::index::DocumentStructure docStruct;
  std::vector<std::string> result;

  std::vector<std::pair<lemur::api::DOCID_T, int> > docSorted;
  for( size_t i=0; i<documentIDs.size(); i++ ) {
    docSorted.push_back( std::make_pair( documentIDs[i], i ) );
  }
  std::sort( docSorted.begin(), docSorted.end() );

  for( size_t i=0; i<docSorted.size(); i++ ) {
    vank::collection::Repository::index_state indexes = _repository.indexes();
    bool docStructLoaded = true;
    lemur::api::DOCID_T documentID = docSorted[i].first;
    if ( documentID != lastDoc ) {
      vank::index::Index * index = _indexWithDocument(indexes, documentID);
      const vank::index::TermList * termList = index->termList( documentID );
      if ( termList != 0 ) {
        docStruct.setIndex( *index );
        docStruct.loadStructure( termList->fields() );
        delete termList;
        lastDoc = docStructLoaded;
      } else {
        docStructLoaded = false;
      }
    }

    std::string path = "";
    if ( docStructLoaded ) {
      path = docStruct.path( docStruct.findLeaf( pathBegins[docSorted[i].second],
                                                 pathEnds[docSorted[i].second] ) );
    }
    result.push_back( path );
  }

  std::vector<std::string> actual;
  actual.resize( documentIDs.size() );
  for( size_t i=0; i<docSorted.size(); i++ ) {
    actual[docSorted[i].second] = result[i];
  }

  return new vank::server::LocalQueryServerMetadataResponse( actual );
}

//
// setMaxWildcardTerms
//
void vank::server::LocalQueryServer::setMaxWildcardTerms(int maxTerms) {
  _maxWildcardMatchesPerTerm = maxTerms;
}
