/*==========================================================================
 * Copyright (c) 2004 University of Massachusetts.  All Rights Reserved.
 *
 * Use of the Lemur Toolkit for Language Modeling and Information Retrieval
 * is subject to the terms of the software license set forth in the LICENSE
 * file included with this software, and also available at
 * http://www.lemurproject.org/license.html
 *
 *==========================================================================
 */


//
// NetworkServerProxy
//
// 20 May 2004 -- tds
//

#include "vank/NetworkServerProxy.hpp"
#include "vank/ParsedDocument.hpp"
#include "vank/ScopedLock.hpp"
#include <iostream>

namespace vank
{
  namespace server
  {

    //
    // NetworkServerProxyResponse
    //
    class NetworkServerProxyResponse : public QueryServerResponse {
    private:
      vank::net::QueryResponseUnpacker _unpacker;
      vank::net::NetworkMessageStream* _stream;

    public:
      NetworkServerProxyResponse( vank::net::NetworkMessageStream* stream ) :
        _unpacker( stream ),
        _stream( stream )
      {
      }

      ~NetworkServerProxyResponse() {
        _stream->mutex().unlock();
      }

      vank::infnet::InferenceNetwork::MAllResults& getResults() {
        return _unpacker.getResults();
      }
    };

    //
    // NetworkServerProxyDocumentsResponse
    //

    class NetworkServerProxyDocumentsResponse : public QueryServerDocumentsResponse {
    private:
      vank::net::NetworkMessageStream* _stream;
      std::vector<vank::api::ParsedDocument*> _documents;

    public:
      NetworkServerProxyDocumentsResponse( vank::net::NetworkMessageStream* stream ) :
        _stream(stream)
      {
      }

      ~NetworkServerProxyDocumentsResponse() {
        _stream->mutex().unlock();
      }

      // caller deletes the ParsedDocuments
      std::vector<vank::api::ParsedDocument*>& getResults() {
        vank::net::XMLReplyReceiver r;
        r.wait(_stream);

        if( r.getReply() ) {
          vank::xml::XMLNode* reply = r.getReply();
          size_t numChildren = reply->getChildren().size();

          for( size_t i=0; i<numChildren; i++ ) {
            vank::utility::Buffer buffer;
            int textOffset = 0;
            vank::utility::greedy_vector<int> metadataKeyOffset;
            vank::utility::greedy_vector<int> metadataValueOffset;

            const vank::xml::XMLNode* child = reply->getChildren()[i];
            const vank::xml::XMLNode* metadata = child->getChild( "metadata" );

            // allocate room for the ParsedDocument
            buffer.write( sizeof(vank::api::ParsedDocument) );

            if( metadata ) {
              for( size_t j=0; j<metadata->getChildren().size(); j++ ) {
                std::string key = metadata->getChildren()[j]->getChildValue("key");
                std::string value = metadata->getChildren()[j]->getChildValue("value");

                metadataKeyOffset.push_back( (int)buffer.position() );
                strcpy( buffer.write( key.size()+1 ), key.c_str() );

                metadataValueOffset.push_back( (int)buffer.position() );
                int length = base64_decode( buffer.write( value.size() ), (int)value.size(), value );
                buffer.unwrite( value.size()-length );
              }
            }

            const vank::xml::XMLNode* textNode = child->getChild("text");
            std::string text;

            if( textNode ) {
              text = textNode->getValue();
              textOffset = (int)buffer.position();
              int length = base64_decode( buffer.write( text.size() ), (int)text.size(), text );
              buffer.unwrite( text.size()-length );
            }

            INT64 contentOffset = 0;
            INT64 contentLength = 0;
            const vank::xml::XMLNode* contentNode = child->getChild("content");
            if (contentNode) {
              contentOffset = string_to_i64(contentNode->getValue());
            }
            const vank::xml::XMLNode* contentLengthNode = child->getChild("contentLength");
            if (contentLengthNode) {
              contentLength = string_to_i64(contentLengthNode->getValue());
            }

            // now all of our data is in the buffer, so we can allocate a return structure
            new(buffer.front()) vank::api::ParsedDocument;
            vank::api::ParsedDocument* parsedDocument = (vank::api::ParsedDocument*) buffer.front();

            const vank::xml::XMLNode* positions = child->getChild( "positions" );

            if( positions ) {
              const std::vector<vank::xml::XMLNode*>& children = positions->getChildren();


              for( size_t j=0; j<children.size(); j++ ) {
                vank::parse::TermExtent extent;
                std::string begin = children[j]->getChildValue("begin");
                std::string end = children[j]->getChildValue("end");

                extent.begin = (int) string_to_i64( begin );
                extent.end = (int) string_to_i64( end );

                parsedDocument->positions.push_back( extent );
              }
            }

            for( size_t j=0; j<metadataKeyOffset.size(); j++ ) {
              vank::parse::MetadataPair pair;

              pair.key = buffer.front() + metadataKeyOffset[j];
              pair.value = buffer.front() + metadataValueOffset[j];

              if( metadataKeyOffset.size() > j+1 )
                pair.valueLength = metadataKeyOffset[j+1] - metadataValueOffset[j];
              else
                pair.valueLength = textOffset;

              parsedDocument->metadata.push_back( pair );
            }

            parsedDocument->text = buffer.front() + textOffset;
            parsedDocument->textLength = buffer.position() - textOffset;
            parsedDocument->content = parsedDocument->text + contentOffset;
            parsedDocument->contentLength = contentLength;
            buffer.detach();

            _documents.push_back( parsedDocument );
          }
        }

        return _documents;
      }
    };

    //
    // NetworkServerProxyMetadataResponse
    //

    class NetworkServerProxyMetadataResponse : public QueryServerMetadataResponse {
    private:
      std::vector<std::string> _metadata;
      vank::net::NetworkMessageStream* _stream;

    public:
      NetworkServerProxyMetadataResponse( vank::net::NetworkMessageStream* stream ) :
        _stream(stream)
      {
      }

      ~NetworkServerProxyMetadataResponse() {
        _stream->mutex().unlock();
      }

      std::vector<std::string>& getResults() {
        vank::net::XMLReplyReceiver r;
        r.wait(_stream);

        // parse the result
        vank::xml::XMLNode* reply = r.getReply();
        vank::utility::Buffer metadataBuffer;

        for( size_t i=0; i<reply->getChildren().size(); i++ ) {
          const vank::xml::XMLNode* meta = reply->getChildren()[i];
          const std::string& input = meta->getValue();

          std::string value;
          base64_decode_string( value, input );
          _metadata.push_back(value);

          metadataBuffer.clear();
        }

        return _metadata;
      }
    };

    //
    // NetworkServerProxyVectorsResponse
    //

    class NetworkServerProxyVectorsResponse : public QueryServerVectorsResponse {
    public:
      std::vector<vank::api::DocumentVector*> _vectors;
      vank::net::NetworkMessageStream* _stream;
      bool _readResponse;

    public:
      NetworkServerProxyVectorsResponse( vank::net::NetworkMessageStream* stream )
        :
        _stream(stream),
        _readResponse(false)
      {
      }

      ~NetworkServerProxyVectorsResponse() {
        _stream->mutex().unlock();
      }

      std::vector<vank::api::DocumentVector*>& getResults() {
        if( !_readResponse ) {
          vank::net::XMLReplyReceiver r;
          r.wait( _stream );

          const vank::xml::XMLNode* reply = r.getReply();
          const std::vector<vank::xml::XMLNode*>& children = reply->getChildren();

          for( size_t i=0; i<children.size(); i++ ) {
            const vank::xml::XMLNode* stems = children[i]->getChild("stems");
            const vank::xml::XMLNode* positions = children[i]->getChild("positions");
            const vank::xml::XMLNode* fields = children[i]->getChild("fields");

            vank::api::DocumentVector* result = new vank::api::DocumentVector;

            for( size_t j=0; j<stems->getChildren().size(); j++ ) {
              // have to use base64 coding, in case the stem contains '<', '>', etc.
              std::string stem;
              base64_decode_string(stem, stems->getChildren()[j]->getValue());
              result->stems().push_back( stem );
            }

            std::vector<int>& positionsVector = result->positions();

            for( size_t j=0; j<positions->getChildren().size(); j++ ) {
              const std::string& stringText = positions->getChildren()[j]->getValue();
              INT64 position = string_to_i64( stringText );
              positionsVector.push_back( int(position) );
            }

            std::vector<vank::api::DocumentVector::Field>& fieldVector = result->fields();

            for( size_t j=0; j<fields->getChildren().size(); j++ ) {
              const vank::xml::XMLNode* field = fields->getChildren()[j];

              const vank::xml::XMLNode* nameField = field->getChild("name");
              const vank::xml::XMLNode* numberField = field->getChild("number");
              const vank::xml::XMLNode* beginField = field->getChild("begin");
              const vank::xml::XMLNode* endField = field->getChild("end");
              const vank::xml::XMLNode* ordinalField = field->getChild("ordinal");
              const vank::xml::XMLNode* pOrdinalField = field->getChild("parentOrdinal");

              vank::api::DocumentVector::Field f;

              f.name = nameField->getValue();
              f.number = string_to_i64( numberField->getValue() );
              f.begin = int(string_to_i64( beginField->getValue() ));
              f.end = int(string_to_i64( endField->getValue() ));
              f.ordinal = int(string_to_i64( ordinalField->getValue() ));
              f.parentOrdinal = int(string_to_i64( pOrdinalField->getValue() ));

              fieldVector.push_back(f);
            }

            _vectors.push_back(result);
          }

          _readResponse = true;
        }

        return _vectors;
      }
    };

    //
    // NetworkServerProxyDocumentIDsResponse
    //

    class NetworkServerProxyDocumentIDsResponse : public QueryServerDocumentIDsResponse {
    public:
      std::vector<lemur::api::DOCID_T> _documentIDs;
      vank::net::NetworkMessageStream* _stream;
      bool _readResponse;

    public:
      NetworkServerProxyDocumentIDsResponse( vank::net::NetworkMessageStream* stream )
        :
        _stream(stream),
        _readResponse(false)
      {
      }

      ~NetworkServerProxyDocumentIDsResponse() {
        _stream->mutex().unlock();
      }

      std::vector<lemur::api::DOCID_T>& getResults() {
        if( !_readResponse ) {
          vank::net::XMLReplyReceiver r;
          r.wait( _stream );

          const vank::xml::XMLNode* reply = r.getReply();
          const std::vector<vank::xml::XMLNode*>& children = reply->getChildren();

          for( size_t i=0; i<children.size(); i++ ) {
            vank::xml::XMLNode* child = children[i];
            _documentIDs.push_back( string_to_i64( child->getValue() ) );
          }
        }

        return _documentIDs;
      }
    };
  }
}

//
// NetworkServerProxy code
//


vank::server::NetworkServerProxy::NetworkServerProxy( vank::net::NetworkMessageStream* stream ) :
  _stream(stream)
{
}

//
// _numericRequest
//
// Sends a request for a numeric quantity; deletes the node
// passed in as a parameter
//

INT64 vank::server::NetworkServerProxy::_numericRequest( vank::xml::XMLNode* node ) {
  vank::thread::ScopedLock lock( _stream->mutex() );
  _stream->request( node );
  delete node;

  vank::net::XMLReplyReceiver r;
  r.wait( _stream );

  vank::xml::XMLNode* reply = r.getReply();
  return string_to_i64( reply->getValue() );
}

//
// _stringRequest
//
// Sends a request for a string quantity; deletes the node
// passed in as a parameter
//

std::string vank::server::NetworkServerProxy::_stringRequest( vank::xml::XMLNode* node ) {
  vank::thread::ScopedLock lock( _stream->mutex() );
  _stream->request( node );
  delete node;

  vank::net::XMLReplyReceiver r;
  r.wait( _stream );

  vank::xml::XMLNode* reply = r.getReply();
  return reply->getValue();
}

//
// runQuery
//

vank::server::QueryServerResponse* vank::server::NetworkServerProxy::runQuery( std::vector<vank::lang::Node*>& roots, int resultsRequested, bool optimize ) {
  vank::lang::Packer packer;

  for( size_t i=0; i<roots.size(); i++ ) {
    packer.pack( roots[i] );
  }

  vank::xml::XMLNode* query = packer.xml();
  query->addAttribute( "resultsRequested", i64_to_string(resultsRequested) );
  query->addAttribute( "optimize", optimize ? "1" : "0" );

  _stream->mutex().lock();
  _stream->request( query );

  return new vank::server::NetworkServerProxyResponse( _stream );
}

//
// documentMetadata
//

vank::server::QueryServerMetadataResponse* vank::server::NetworkServerProxy::documentMetadata( const std::vector<lemur::api::DOCID_T>& documentIDs, const std::string& attributeName ) {
  vank::xml::XMLNode* request = new vank::xml::XMLNode( "document-metadata" );
  vank::xml::XMLNode* field = new vank::xml::XMLNode( "field", attributeName );
  vank::xml::XMLNode* documents = new vank::xml::XMLNode( "documents" );

  // build request
  for( size_t i=0; i<documentIDs.size(); i++ ) {
    documents->addChild( new vank::xml::XMLNode( "document", i64_to_string( documentIDs[i] ) ) );
  }
  request->addChild( field );
  request->addChild( documents );

  // send request
  _stream->mutex().lock();
  _stream->request( request );
  delete request;

  return new vank::server::NetworkServerProxyMetadataResponse( _stream );
}

//
// pathNames
//

vank::server::QueryServerMetadataResponse* vank::server::NetworkServerProxy::pathNames( const std::vector<lemur::api::DOCID_T>& documentIDs, const std::vector<int>& begins, const std::vector<int>& ends ) {
  vank::xml::XMLNode* request = new vank::xml::XMLNode( "path-names" );
  vank::xml::XMLNode* documents = new vank::xml::XMLNode( "paths" );

  // build request
  for( size_t i=0; i<documentIDs.size(); i++ ) {
    documents->addChild( new vank::xml::XMLNode( "document", i64_to_string( documentIDs[i] ) ) );
    documents->addChild( new vank::xml::XMLNode( "begin", i64_to_string( begins[i] ) ) );
    documents->addChild( new vank::xml::XMLNode( "end", i64_to_string( ends[i] ) ) );
  }
  request->addChild( documents );

  // send request
  _stream->mutex().lock();
  _stream->request( request );
  delete request;

  return new vank::server::NetworkServerProxyMetadataResponse( _stream );
}

//
// documents
//

vank::server::QueryServerDocumentsResponse* vank::server::NetworkServerProxy::documents( const std::vector<lemur::api::DOCID_T>& documentIDs ) {
  vank::xml::XMLNode* docRequest = new vank::xml::XMLNode( "documents" );

  for( size_t i=0; i<documentIDs.size(); i++ ) {
    docRequest->addChild( new vank::xml::XMLNode("doc", i64_to_string(documentIDs[i])) );
  }

  _stream->mutex().lock();
  _stream->request( docRequest );
  delete docRequest;

  return new vank::server::NetworkServerProxyDocumentsResponse( _stream );
}

//
// documentsFromMetadata
//

vank::server::QueryServerDocumentsResponse* vank::server::NetworkServerProxy::documentsFromMetadata( const std::string& attributeName, const std::vector<std::string>& attributeValues ) {
  vank::xml::XMLNode* docRequest = new vank::xml::XMLNode( "documents-from-metadata" );

  // store the attribute name
  docRequest->addChild( new vank::xml::XMLNode( "attributeName", attributeName ) );

  // serialize the attributeValues (in a tree called "attributeValues", with members called "attributeValue")
  vank::xml::XMLNode* attributeValuesNode = new vank::xml::XMLNode( "attributeValues" );
  for( size_t i=0; i<attributeValues.size(); i++ ) {
    attributeValuesNode->addChild( new vank::xml::XMLNode( "attributeValue", attributeValues[i] ) );
  }
  docRequest->addChild( attributeValuesNode );

  // request the documents
  _stream->mutex().lock();
  _stream->request( docRequest );
  delete docRequest;

  return new vank::server::NetworkServerProxyDocumentsResponse( _stream );
}

//
// documentIDsFromMetadata
//

vank::server::QueryServerDocumentIDsResponse* vank::server::NetworkServerProxy::documentIDsFromMetadata( const std::string& attributeName, const std::vector<std::string>& attributeValues ) {
  vank::xml::XMLNode* docRequest = new vank::xml::XMLNode( "docids-from-metadata" );

  // store the attribute name
  docRequest->addChild( new vank::xml::XMLNode( "attributeName", attributeName ) );

  // serialize the attributeValues (in a tree called "attributeValues", with members called "attributeValue")
  vank::xml::XMLNode* attributeValuesNode = new vank::xml::XMLNode( "attributeValues" );
  for( size_t i=0; i<attributeValues.size(); i++ ) {
    attributeValuesNode->addChild( new vank::xml::XMLNode( "attributeValue", attributeValues[i] ) );
  }
  docRequest->addChild( attributeValuesNode );

  // request the documents
  _stream->mutex().lock();
  _stream->request( docRequest );
  delete docRequest;

  return new vank::server::NetworkServerProxyDocumentIDsResponse( _stream );
}

//
// termCount
//

INT64 vank::server::NetworkServerProxy::termCount() {
  //  std::auto_ptr<vank::xml::XMLNode> request( new vank::xml::XMLNode( "term-count" ) );
  vank::xml::XMLNode *request = new vank::xml::XMLNode( "term-count" ) ;
  return _numericRequest( request );
}

//
// termCount
//

INT64 vank::server::NetworkServerProxy::termCount( const std::string& term ) {
  vank::xml::XMLNode* request = new vank::xml::XMLNode( "term-count-text", term );
  return _numericRequest( request );
}

INT64 vank::server::NetworkServerProxy::termCountUnique( ) {
  vank::xml::XMLNode* request = new vank::xml::XMLNode( "term-count-unique" );
  return _numericRequest( request );
}

INT64 vank::server::NetworkServerProxy::stemCount( const std::string& term ) {
  vank::xml::XMLNode* request = new vank::xml::XMLNode( "stem-count-text", term );
  return _numericRequest( request );
}

std::string vank::server::NetworkServerProxy::termName( lemur::api::TERMID_T term ) {
  vank::xml::XMLNode* request = new vank::xml::XMLNode( "term-name", i64_to_string(term) );
  return _stringRequest( request );
}

std::string vank::server::NetworkServerProxy::stemTerm( const std::string & term ) {
  vank::xml::XMLNode* request = new vank::xml::XMLNode( "term-stem", term );
  return _stringRequest( request );
}

lemur::api::TERMID_T vank::server::NetworkServerProxy::termID( const std::string& term ) {
  vank::xml::XMLNode* request = new vank::xml::XMLNode( "term-id", term );
  return _numericRequest( request );
}

INT64 vank::server::NetworkServerProxy::termFieldCount( const std::string& term, const std::string& field ) {
  vank::xml::XMLNode* request = new vank::xml::XMLNode( "term-field-count" );
  vank::xml::XMLNode* termNode = new vank::xml::XMLNode( "term-text", term );
  vank::xml::XMLNode* fieldNode = new vank::xml::XMLNode( "field", field );
  request->addChild( termNode );
  request->addChild( fieldNode );

  return _numericRequest( request );
}

INT64 vank::server::NetworkServerProxy::stemFieldCount( const std::string& stem, const std::string& field ) {
  vank::xml::XMLNode* request = new vank::xml::XMLNode( "stem-field-count" );
  vank::xml::XMLNode* stemNode = new vank::xml::XMLNode( "stem-text", stem );
  vank::xml::XMLNode* fieldNode = new vank::xml::XMLNode( "field", field );
  request->addChild( stemNode );
  request->addChild( fieldNode );

  return _numericRequest( request );
}

std::vector<std::string> vank::server::NetworkServerProxy::fieldList() {
  // std::auto_ptr<vank::xml::XMLNode> request( new vank::xml::XMLNode( "field-list" ) );
  std::unique_ptr<vank::xml::XMLNode> request( new vank::xml::XMLNode( "field-list" ) );
  vank::thread::ScopedLock( _stream->mutex() );
  _stream->request( request.get() );

  vank::net::XMLReplyReceiver r;
  r.wait( _stream );

  vank::xml::XMLNode* reply = r.getReply();
  const std::vector<vank::xml::XMLNode*>& children = reply->getChildren();
  std::vector<std::string> result;

  for( size_t i=0; i<children.size(); i++ ) {
    result.push_back( children[i]->getValue() );
  }

  return result;
}

int vank::server::NetworkServerProxy::documentLength( lemur::api::DOCID_T documentID ) {
  vank::xml::XMLNode* request = new vank::xml::XMLNode( "document-length", i64_to_string(documentID) );
  return (int) _numericRequest( request );
}

INT64 vank::server::NetworkServerProxy::documentCount() {
  vank::xml::XMLNode* request = new vank::xml::XMLNode( "document-count" );
  return _numericRequest( request );
}

INT64 vank::server::NetworkServerProxy::documentCount( const std::string& term ) {
  vank::xml::XMLNode* request = new vank::xml::XMLNode( "document-term-count", term );
  return _numericRequest( request );
}

INT64 vank::server::NetworkServerProxy::documentStemCount( const std::string& term ) {
  vank::xml::XMLNode* request = new vank::xml::XMLNode( "document-stem-count", term );
  return _numericRequest( request );
}

vank::server::QueryServerVectorsResponse* vank::server::NetworkServerProxy::documentVectors( const std::vector<lemur::api::DOCID_T>& documentIDs ) {
  vank::xml::XMLNode* request = new vank::xml::XMLNode( "document-vectors" );

  for( size_t i=0; i<documentIDs.size(); i++ ) {
    request->addChild( new vank::xml::XMLNode( "document", i64_to_string(documentIDs[i]) ) );
  }

  _stream->mutex().lock();
  _stream->request( request );
  delete request;

  return new vank::server::NetworkServerProxyVectorsResponse( _stream );
}

//
// setMaxWildcardTerms
//
void vank::server::NetworkServerProxy::setMaxWildcardTerms(int maxTerms) {
  vank::xml::XMLNode* request = new vank::xml::XMLNode( "max-wildcard-terms", i64_to_string(maxTerms) );
  INT64 termMax=_numericRequest( request );
}
