/*==========================================================================
 * Copyright (c) 2004 University of Massachusetts.  All Rights Reserved.
 *
 * Use of the Lemur Toolkit for Language Modeling and Information Retrieval
 * is subject to the terms of the software license set forth in the LICENSE
 * file included with this software, and also available at
 * http://www.lemurproject.org/license.html
 *
 *==========================================================================
 */


//
// OrNode
//
// 6 July 2004 -- tds
//

#include "vank/OrNode.hpp"
#include "lemur/lemur-compat.hpp"
#include "vank/ScoredExtentResult.hpp"
#include "vank/Annotator.hpp"

vank::infnet::OrNode::OrNode( const std::string& name ) :
  _name(name)
{
}

vank::infnet::OrNode::OrNode( const std::string& name, const std::vector<BeliefNode*>& children ) :
  _children( children ),
  _name( name )
{
  // if we have children - set the sibling flag...
  if (_children.size() > 1) {
    for (int i=0; i < _children.size(); i++) {
      if (_children[i]) {
        _children[i]->setSiblingsFlag(1);
      }
    }
  }
}

void vank::infnet::OrNode::setSiblingsFlag(int f){
  // set flag for child nodes
  for(int i=0;i<_children.size();i++) {
    _children[i]->setSiblingsFlag(f);
  }
}


const vank::utility::greedy_vector<vank::api::ScoredExtentResult>& vank::infnet::OrNode::score( lemur::api::DOCID_T documentID, vank::index::Extent &extent, int documentLength ) {
  double notScore = 1;

  for( size_t i=0; i<_children.size(); i++ ) {
    const vank::utility::greedy_vector<vank::api::ScoredExtentResult>& childResults = _children[i]->score( documentID, extent, documentLength );

    for( size_t j=0; j<childResults.size(); j++ ) {
      notScore *= 1. - exp( childResults[j].score );
    }
  }

  double score = log( 1. - notScore );
  _scores.clear();
  vank::api::ScoredExtentResult result(extent);
  result.score=score;
  result.document=documentID;
  _scores.push_back( result );

  return _scores;
}

void vank::infnet::OrNode::annotate( class vank::infnet::Annotator& annotator, lemur::api::DOCID_T documentID, vank::index::Extent &extent ) {
  annotator.add(this, documentID, extent);

  for( size_t i=0; i<_children.size(); i++ ) {
    _children[i]->annotate( annotator, documentID, extent );
  }
}

double vank::infnet::OrNode::maximumScore() {
  double notScore = 1;

  for( size_t i=0; i<_children.size(); i++ ) {
    notScore *= 1. - exp( _children[i]->maximumBackgroundScore() );
  }

  return log( 1. - notScore );
}

double vank::infnet::OrNode::maximumBackgroundScore() {
  double notScore = 1;

  for( size_t i=0; i<_children.size(); i++ ) {
    notScore *= 1. - exp( _children[i]->maximumScore() );
  }

  return log( 1. - notScore );
}

bool vank::infnet::OrNode::hasMatch( lemur::api::DOCID_T documentID ) {
  for( size_t i=0; i<_children.size(); i++ ) {
    if( _children[i]->hasMatch( documentID ) )
      return true;
  }

  return false;
}

const vank::utility::greedy_vector<bool>& vank::infnet::OrNode::hasMatch( lemur::api::DOCID_T documentID, const vank::utility::greedy_vector<vank::index::Extent>& extents ) {
  _matches.clear();
  _matches.resize( extents.size(), false );

  for( size_t i=0; i<_children.size(); i++ ) {
    const vank::utility::greedy_vector<bool>& kidMatches = _children[i]->hasMatch( documentID, extents );

    for( size_t j=0; j<kidMatches.size(); j++ ) {
      if( kidMatches[j] ) {
        _matches[j] = true;
      }
    }
  }

  return _matches;
}

lemur::api::DOCID_T vank::infnet::OrNode::nextCandidateDocument() {
  lemur::api::DOCID_T nextCandidate = MAX_INT32;

  for( size_t i=0; i<_children.size(); i++ ) {
    nextCandidate = lemur_compat::min<lemur::api::DOCID_T>( nextCandidate, _children[i]->nextCandidateDocument() );
  }

  return nextCandidate;
}

void vank::infnet::OrNode::indexChanged( vank::index::Index& index ) {
  // do nothing
}

const std::string& vank::infnet::OrNode::getName() const {
  return _name;
}
