/*==========================================================================
 * Copyright (c) 2004 University of Massachusetts.  All Rights Reserved.
 *
 * Use of the Lemur Toolkit for Language Modeling and Information Retrieval
 * is subject to the terms of the software license set forth in the LICENSE
 * file included with this software, and also available at
 * http://www.lemurproject.org/license.html
 *
 *==========================================================================
 */

#include "vank/TermFrequencyBeliefNode.hpp"
#include "vank/InferenceNetwork.hpp"
#include <math.h>

vank::infnet::TermFrequencyBeliefNode::TermFrequencyBeliefNode( const std::string& name,
                                                                 class InferenceNetwork& network,
                                                                 int listID,
                                                                 vank::query::TermScoreFunction& scoreFunction )
  :
  _name(name),
  _network(network),
  _listID(listID),
  _function(scoreFunction)
{
  _maximumBackgroundScore = VANK_HUGE_SCORE;
  _maximumScore = VANK_HUGE_SCORE;
}

vank::infnet::TermFrequencyBeliefNode::~TermFrequencyBeliefNode() {
}

const vank::utility::greedy_vector<vank::index::DocListIterator::TopDocument>& vank::infnet::TermFrequencyBeliefNode::topdocs() const {
  if( _list )
    return _list->topDocuments();

  return _emptyTopdocs;
}

lemur::api::DOCID_T vank::infnet::TermFrequencyBeliefNode::nextCandidateDocument() {
  if( _list ) {
    const vank::index::DocListIterator::DocumentData* entry = _list->currentEntry();

    if( entry ) {
      return entry->document;
    }
  }

  return MAX_INT32;
}

double vank::infnet::TermFrequencyBeliefNode::maximumBackgroundScore() {
  return _maximumBackgroundScore;
}

double vank::infnet::TermFrequencyBeliefNode::maximumScore() {
  return _maximumScore;
}

const vank::utility::greedy_vector<vank::api::ScoredExtentResult>& vank::infnet::TermFrequencyBeliefNode::score( lemur::api::DOCID_T documentID, vank::index::Extent &extent, int documentLength ) {
  assert( extent.begin == 0 && extent.end == documentLength ); // FrequencyListCopier ensures this condition
  _extents.clear();

  double score = 0;

  if( _list ) {
    const vank::index::DocListIterator::DocumentData* entry = _list->currentEntry();
    int count = ( entry && entry->document == documentID ) ? (int)entry->positions.size() : 0;
    score = _function.scoreOccurrence( count, documentLength );

    assert( score <= _maximumScore || _list->topDocuments().size() > 0 );
    assert( score <= _maximumBackgroundScore || count != 0 );
  } else {
    score = _function.scoreOccurrence( 0, documentLength );
  }

  vank::api::ScoredExtentResult result(extent);
  result.score=score;
  result.document=documentID;
  _extents.push_back( result );

  return _extents;
}

bool vank::infnet::TermFrequencyBeliefNode::hasMatch( lemur::api::DOCID_T documentID ) {
  if( _list ) {
    const vank::index::DocListIterator::DocumentData* entry = _list->currentEntry();
    return ( entry && entry->document == documentID );
  }

  return false;
}

const vank::utility::greedy_vector<bool>& vank::infnet::TermFrequencyBeliefNode::hasMatch( lemur::api::DOCID_T documentID, const vank::utility::greedy_vector<vank::index::Extent>& extents ) {
  assert( false && "A TermFrequencyBeliefNode should never be asked for position information" );

  _matches.resize( extents.size(), false );
  return _matches;
}

const std::string& vank::infnet::TermFrequencyBeliefNode::getName() const {
  return _name;
}

void vank::infnet::TermFrequencyBeliefNode::indexChanged( vank::index::Index& index ) {
  // fetch the next inverted list
  _list = _network.getDocIterator( _listID );

  if( !_list ) {
    _maximumBackgroundScore = VANK_HUGE_SCORE;
    _maximumScore = VANK_HUGE_SCORE;
  } else {
    // maximum fraction
    double maximumFraction = 1;

    if( _list->topDocuments().size() ) {
      if( vank::api::Parameters::instance().get( "topdocs", true ) == false ) {
        //        std::cout << "using no topdocs!" << std::endl;
        const vank::index::DocListIterator::TopDocument& document = _list->topDocuments().front();
        maximumFraction = double(document.count) / double(document.length);
      } else {
      const vank::index::DocListIterator::TopDocument& document = _list->topDocuments().back();
      maximumFraction = double(document.count) / double(document.length);
      }
    }

    vank::index::TermData* termData = _list->termData();

    double maxOccurrences = ceil( double(termData->maxDocumentLength) * maximumFraction );

    _maximumScore = _function.scoreOccurrence( maxOccurrences, termData->maxDocumentLength );
    _maximumBackgroundScore = _function.scoreOccurrence( 0, 1 );
  }
}

void vank::infnet::TermFrequencyBeliefNode::annotate( vank::infnet::Annotator& annotator, lemur::api::DOCID_T documentID, vank::index::Extent &extent ) {
  // can't annotate -- don't have position info
}
