/*==========================================================================
 * Copyright (c) 2004 University of Massachusetts.  All Rights Reserved.
 *
 * Use of the Lemur Toolkit for Language Modeling and Information Retrieval
 * is subject to the terms of the software license set forth in the LICENSE
 * file included with this software, and also available at
 * http://www.lemurproject.org/license.html
 *
 *==========================================================================
 */


//
// UTF8Transcoder
//
// 27 June 2006 -- refactor to put the code in a separate compilation unit.

#include "vank/UTF8Transcoder.hpp"

int vank::parse::UTF8Transcoder::_count_bytes( unsigned char* buf, int index, int max_index, int how_many ) {

  int count = 0;

  for ( int i = index + 1; i <= index + how_many; i++ )
    if ( i < max_index && buf[i] && buf[i] != '\0' &&
         0x80 <= buf[i] && buf[i] <= 0xBF )
      count++;

  return count;
}

void vank::parse::UTF8Transcoder::store_interval( vank::utility::HashTable<UINT64,const int>&
                                                   table, UINT64 start, UINT64 end,
                                                   const int cls ) {

  for ( UINT64 i = start; i <= end; i++ )
    table.insert( i, cls );
}


vank::parse::UTF8Transcoder::UTF8Transcoder() { _initHT();}
vank::parse::UTF8Transcoder::~UTF8Transcoder() {}

void vank::parse::UTF8Transcoder::utf8_encode( UINT64 code, char* buf, int* octets ) {

  int bytes = 0;

  if ( code <= 0x7F ) {

    if ( buf ) {

      buf[0] = (char)code;
      buf[1] = '\0';
    }

    bytes = 1;

  } else if ( code >= 0x80 && code <= 0x7FF ) {

    if ( buf ) {

      buf[0] = 0xC0 | ( 0x1F & ( code >> 6 ) );
      buf[1] = 0x80 | ( 0x3F & code );
      buf[2] = '\0';
    }

    bytes = 2;

  } else if ( code >= 0x800 && code <= 0xFFFF ) {

    if ( buf ) {

      buf[0] = 0xE0 | ( 0xF & ( code >> 12 ) );
      buf[1] = 0x80 | ( 0x3F & ( code >> 6 ) );
      buf[2] = 0x80 | ( 0x3F & code );
      buf[3] = '\0';
    }

    bytes = 3;

  } else if ( code >= 0x1000 && code <= 0x1FFFFF ) {

    if ( buf ) {

      buf[0] = 0xF0 | ( 0x7 & ( code >> 18 ) );
      buf[1] = 0x80 | ( 0x3F & ( code >> 12 ) );
      buf[2] = 0x80 | ( 0x3F & ( code >> 6 ) );
      buf[3] = 0x80 | ( 0x3F & code );
      buf[4] = '\0';
    }

    bytes = 4;

  } else if ( code >= 0x200000 && code <= 0x3FFFFFF ) {

    if ( buf ) {

      buf[0] = 0xF8 | ( 0x7 & ( code >> 24 ) );
      buf[1] = 0x80 | ( 0x3F & ( code >> 18 ) );
      buf[2] = 0x80 | ( 0x3F & ( code >> 12 ) );
      buf[3] = 0x80 | ( 0x3F & ( code >> 6 ) );
      buf[4] = 0x80 | ( 0x3F & code );
      buf[5] = '\0';
    }

    bytes = 5;

  } else if ( code >= 0x4000000 && code <= 0x7FFFFFFF ) {

    if ( buf ) {

      buf[0] = 0xFC | ( 0x1 & ( code >> 30 ) );
      buf[1] = 0x80 | ( 0x3F & ( code >> 24 ) );
      buf[2] = 0x80 | ( 0x3F & ( code >> 18 ) );
      buf[3] = 0x80 | ( 0x3F & ( code >> 12 ) );
      buf[4] = 0x80 | ( 0x3F & ( code >> 6 ) );
      buf[5] = 0x80 | ( 0x3F & code );
      buf[6] = '\0';
    }

    bytes = 5;

  } else {

    if ( buf ) buf[0] = '\0';
  }

  if ( octets ) *octets = bytes;
}

void vank::parse::UTF8Transcoder::utf8_decode( const char* buf_in, UINT64** codes, int* characters,
                                                int* malformed, int** offsets, int** lengths ) {

  int len = strlen( buf_in );
  int i = 0; // index into codes array
  int j = 0; // index into input buf
  UINT64 code = 0;
  int bytes;

  unsigned char* buf = (unsigned char*)buf_in;

  while ( j < len ) {

    bool code_ok = false;
    bytes = 0;

    if ( buf[j] <= 0x7F ) { // 7-bit ASCII character (One-byte sequence)

      code = (UINT64)buf[j];
      bytes = 1;
      code_ok = true;

    } else if ( 0xC0 <= buf[j] && buf[j] <= 0xDF ) { // Two-byte sequence

      bytes = 1 + _count_bytes( buf, j, len, 1 );

      if ( bytes == 2 ) {

        code = ( ( 0x1F & buf[j] ) << 6 ) + ( 0x3F & buf[j + 1] );

        if ( code < 0x80 ) { // Overlong sequence detected

          if ( malformed ) *malformed += 2;

        } else {

          code_ok = true;
        }
      } else { // Subsequent bytes of sequence malformed

        bytes = 1;
        if ( malformed ) *malformed++;
      }

    } else if ( 0xE0 <= buf[j] && buf[j] <= 0xEF ) { // Three-byte sequence

      bytes = 1 + _count_bytes( buf, j, len, 2 );

      if ( bytes == 3 ) {

        code = ( ( 0xF & buf[j] ) << 12 ) + ( ( 0x3F & buf[j + 1] ) << 6 ) +
          ( 0x3F & buf[j + 2] );

        if ( code < 0x800 ) { // Overlong sequence detected

          if ( malformed ) *malformed += 3;

        } else {

          code_ok = true;
        }
      } else { // Subsequent bytes of sequence malformed

        bytes = 1;
        if ( malformed ) *malformed += 3;
      }

    } else if ( 0xF0 <= buf[j] && buf[j] <= 0xF7 ) { // Four-byte sequence

      bytes = 1 + _count_bytes( buf, j, len, 3 );

      if ( bytes == 4 ) {

        code = ( ( 0x7 & buf[j] ) << 18 ) + ( ( 0x3F & buf[j + 1] ) << 12 ) +
          ( ( 0x3F & buf[j + 2] ) << 6 ) + ( 0x3F & buf[j + 3] );

        if ( code < 0x10000 ) { // Overlong sequence detected

          if ( malformed ) *malformed += 4;

        } else {

          code_ok = true;
        }
      } else { // Subsequent bytes of sequence malformed

        bytes = 1;
        if ( malformed ) *malformed += 4;
      }

    } else if ( 0xF8 <= buf[j] && buf[j] <= 0xFB ) { // Five-byte sequence

      bytes = 1 + _count_bytes( buf, j, len, 4 );

      if ( bytes == 5 ) {

        code = ( ( 0x3 & buf[j] ) << 24 ) + ( ( 0x3F & buf[j + 1] ) << 18 ) +
          ( ( 0x3F & buf[j + 2] ) << 12 ) + ( ( 0x3F & buf[j + 3] ) << 6 ) +
          ( 0x3F & buf[j + 4] );

        if ( code < 0x200000 ) { // Overlong sequence detected

          if ( malformed ) *malformed += 5;
        } else {

          code_ok = true;
        }
      } else { // Subsequent bytes of sequence malformed

        bytes = 1;
        if ( malformed ) *malformed += 5;
      }

    } else if ( 0xFC <= buf[j] && buf[j] <= 0xFD ) { // Six-byte sequence

      bytes = 1 + _count_bytes( buf, j, len, 5 );

      if ( bytes == 6 ) {

        code = ( ( 0x1 & buf[j] ) << 30 ) + ( ( 0x3F & buf[j + 1] ) << 24 ) +
          ( ( 0x3F & buf[j + 2] ) << 18 ) + ( ( 0x3F & buf[j + 3] ) << 12 ) +
          ( ( 0x3F & buf[j + 4] ) << 6 ) + ( 0x3F & buf[j + 5] );

        if ( code < 0x4000000 ) { // Overlong sequence detected

          if ( malformed ) *malformed += 6;

        } else {

          code_ok = true;
        }
      } else { // Subsequent bytes of sequence malformed

        bytes = 1;
        if ( malformed ) *malformed += 6;
      }

    } else { // Bytes 0xFE and 0xFF are invalid in UTF-8
      // Bytes 0x80-0xBF are invalid as the first byte of a
      // sequence.

      bytes = 1;
      if ( malformed ) *malformed += 1;
    }

    if ( code_ok ) {

      if ( offsets ) (*offsets)[i] = j;
      if ( lengths ) (*lengths)[i] = bytes;
      if ( codes ) (*codes)[i] = code;
      i++;
      if ( characters ) (*characters)++;
    }

    j += bytes;
  }

  // Terminate the arrays:
  if ( offsets ) (*offsets)[i] = 0;
  if ( lengths ) (*lengths)[i] = 0;
  if ( codes ) (*codes)[i] = 0;
}

struct intervalEntry {
  UINT64 start;
  UINT64 end;
  const int cls ;
};


static const struct intervalEntry intervals[] = {
  { 0xFF0C, 0xFF0C, vank::parse::CharClass::thousand },
  { 0x002C, 0x002C, vank::parse::CharClass::thousand },
  { 0x066C, 0x066C, vank::parse::CharClass::thousand },

  { 0x002D, 0x002D, vank::parse::CharClass::hyphen },
  { 0x2010, 0x2011, vank::parse::CharClass::hyphen },
  { 0xFF0D, 0xFF0D, vank::parse::CharClass::hyphen },

  { 0x002E, 0x002E, vank::parse::CharClass::decimal },
  { 0x066B, 0x066B, vank::parse::CharClass::decimal },
  { 0xFF0E, 0xFF0E, vank::parse::CharClass::decimal },

  { 0x2000, 0x200A, vank::parse::CharClass::whitespace },
  { 0x2028, 0x2029, vank::parse::CharClass::whitespace },
  { 0x202F, 0x202F, vank::parse::CharClass::whitespace },
  { 0x205F, 0x205F, vank::parse::CharClass::whitespace },
  { 0x0020, 0x0020, vank::parse::CharClass::whitespace },
  { 0x1680, 0x1680, vank::parse::CharClass::whitespace },
  { 0x180E, 0x180E, vank::parse::CharClass::whitespace },
  { 0x00A0, 0x00A0, vank::parse::CharClass::whitespace },
  { 0x3000, 0x3000, vank::parse::CharClass::whitespace },

  { 0x17D4, 0x17D6, vank::parse::CharClass::punctuation },
  { 0x17D8, 0x17DA, vank::parse::CharClass::punctuation },
  { 0x10A50, 0x10A58, vank::parse::CharClass::punctuation },
  { 0x0021, 0x0023, vank::parse::CharClass::punctuation },
  { 0x0026, 0x0026, vank::parse::CharClass::punctuation },
  { 0x0028, 0x002A, vank::parse::CharClass::punctuation },
  { 0x002F, 0x002F, vank::parse::CharClass::punctuation },
  { 0x003A, 0x003B, vank::parse::CharClass::punctuation },
  { 0x003F, 0x0040, vank::parse::CharClass::punctuation },
  { 0x005B, 0x0060, vank::parse::CharClass::punctuation },
  { 0x007B, 0x007B, vank::parse::CharClass::punctuation },
  { 0x007D, 0x007E, vank::parse::CharClass::punctuation },
  { 0x1039F, 0x1039F, vank::parse::CharClass::punctuation },
  { 0x0E4F, 0x0E4F, vank::parse::CharClass::punctuation },
  { 0x0E5A, 0x0E5B, vank::parse::CharClass::punctuation },
  { 0x0374, 0x0375, vank::parse::CharClass::punctuation },
  { 0x037E, 0x037E, vank::parse::CharClass::punctuation },
  { 0x0384, 0x0385, vank::parse::CharClass::punctuation },
  { 0x0387, 0x0387, vank::parse::CharClass::punctuation },
  { 0x10FB, 0x10FB, vank::parse::CharClass::punctuation },
  { 0x0DF4, 0x0DF4, vank::parse::CharClass::punctuation },
  { 0x1800, 0x180A, vank::parse::CharClass::punctuation },
  { 0x060C, 0x060D, vank::parse::CharClass::punctuation },
  { 0x061B, 0x061B, vank::parse::CharClass::punctuation },
  { 0x061E, 0x061F, vank::parse::CharClass::punctuation },
  { 0x066D, 0x066D, vank::parse::CharClass::punctuation },
  { 0x06D4, 0x06D4, vank::parse::CharClass::punctuation },
  { 0xFD3E, 0xFD3F, vank::parse::CharClass::punctuation },
  { 0xFE10, 0xFE19, vank::parse::CharClass::punctuation },
  { 0xFE30, 0xFE52, vank::parse::CharClass::punctuation },
  { 0xFE54, 0xFE61, vank::parse::CharClass::punctuation },
  { 0xFE63, 0xFE63, vank::parse::CharClass::punctuation },
  { 0xFE68, 0xFE68, vank::parse::CharClass::punctuation },
  { 0xFE6A, 0xFE6B, vank::parse::CharClass::punctuation },
  { 0x02C2, 0x02C5, vank::parse::CharClass::punctuation },
  { 0x02D2, 0x02DF, vank::parse::CharClass::punctuation },
  { 0x02E5, 0x02ED, vank::parse::CharClass::punctuation },
  { 0x02EF, 0x02FF, vank::parse::CharClass::punctuation },
  { 0x0F04, 0x0F12, vank::parse::CharClass::punctuation },
  { 0x0F3A, 0x0F3D, vank::parse::CharClass::punctuation },
  { 0x0F85, 0x0F85, vank::parse::CharClass::punctuation },
  { 0x0FD0, 0x0FD1, vank::parse::CharClass::punctuation },
  { 0x1735, 0x1736, vank::parse::CharClass::punctuation },
  { 0x3001, 0x3003, vank::parse::CharClass::punctuation },
  { 0x3008, 0x3011, vank::parse::CharClass::punctuation },
  { 0x3014, 0x301F, vank::parse::CharClass::punctuation },
  { 0x3030, 0x3030, vank::parse::CharClass::punctuation },
  { 0x303D, 0x303D, vank::parse::CharClass::punctuation },
  { 0x30A0, 0x30A0, vank::parse::CharClass::punctuation },
  { 0x30FB, 0x30FB, vank::parse::CharClass::punctuation },
  { 0xFF65, 0xFF65, vank::parse::CharClass::punctuation },
  { 0x169B, 0x169C, vank::parse::CharClass::punctuation },
  { 0x1361, 0x1368, vank::parse::CharClass::punctuation },
  { 0x0589, 0x058A, vank::parse::CharClass::punctuation },
  { 0x19DE, 0x19DF, vank::parse::CharClass::punctuation },
  { 0x2012, 0x2015, vank::parse::CharClass::punctuation },
  { 0x2018, 0x2018, vank::parse::CharClass::punctuation },
  { 0x201A, 0x201F, vank::parse::CharClass::punctuation },
  { 0x2026, 0x2026, vank::parse::CharClass::punctuation },
  { 0x2039, 0x203A, vank::parse::CharClass::punctuation },
  { 0x203C, 0x2040, vank::parse::CharClass::punctuation },
  { 0x2044, 0x2049, vank::parse::CharClass::punctuation },
  { 0x204F, 0x204F, vank::parse::CharClass::punctuation },
  { 0x2053, 0x2054, vank::parse::CharClass::punctuation },
  { 0x207D, 0x207E, vank::parse::CharClass::punctuation },
  { 0x208D, 0x208E, vank::parse::CharClass::punctuation },
  { 0x2768, 0x2775, vank::parse::CharClass::punctuation },
  { 0x27C5, 0x27C6, vank::parse::CharClass::punctuation },
  { 0x27E6, 0x27EB, vank::parse::CharClass::punctuation },
  { 0x2983, 0x2998, vank::parse::CharClass::punctuation },
  { 0x29D8, 0x29DB, vank::parse::CharClass::punctuation },
  { 0x29FC, 0x29FD, vank::parse::CharClass::punctuation },
  { 0x2E00, 0x2E17, vank::parse::CharClass::punctuation },
  { 0x2E1C, 0x2E1D, vank::parse::CharClass::punctuation },
  { 0xFF01, 0xFF03, vank::parse::CharClass::punctuation },
  { 0xFF06, 0xFF06, vank::parse::CharClass::punctuation },
  { 0xFF08, 0xFF0A, vank::parse::CharClass::punctuation },
  { 0xFF0F, 0xFF0F, vank::parse::CharClass::punctuation },
  { 0xFF1A, 0xFF1B, vank::parse::CharClass::punctuation },
  { 0xFF1F, 0xFF20, vank::parse::CharClass::punctuation },
  { 0xFF3B, 0xFF40, vank::parse::CharClass::punctuation },
  { 0xFF5B, 0xFF64, vank::parse::CharClass::punctuation },
  { 0x2CF9, 0x2CFC, vank::parse::CharClass::punctuation },
  { 0x2CFE, 0x2CFF, vank::parse::CharClass::punctuation },
  { 0x0964, 0x0965, vank::parse::CharClass::punctuation },
  { 0x0970, 0x0970, vank::parse::CharClass::punctuation },
  { 0x0700, 0x070D, vank::parse::CharClass::punctuation },
  { 0x00A1, 0x00A1, vank::parse::CharClass::punctuation },
  { 0x00A8, 0x00A8, vank::parse::CharClass::punctuation },
  { 0x00AB, 0x00AB, vank::parse::CharClass::punctuation },
  { 0x00AF, 0x00AF, vank::parse::CharClass::punctuation },
  { 0x00B4, 0x00B4, vank::parse::CharClass::punctuation },
  { 0x00B7, 0x00B8, vank::parse::CharClass::punctuation },
  { 0x00BB, 0x00BB, vank::parse::CharClass::punctuation },
  { 0x00BF, 0x00BF, vank::parse::CharClass::punctuation },
  { 0x05BE, 0x05BE, vank::parse::CharClass::punctuation },
  { 0x05C0, 0x05C0, vank::parse::CharClass::punctuation },
  { 0x05C3, 0x05C3, vank::parse::CharClass::punctuation },
  { 0x05C6, 0x05C6, vank::parse::CharClass::punctuation },
  { 0x05F3, 0x05F4, vank::parse::CharClass::punctuation },
  { 0x10100, 0x10101, vank::parse::CharClass::punctuation },
  { 0x1A1E, 0x1A1F, vank::parse::CharClass::punctuation },
  { 0x104A, 0x104F, vank::parse::CharClass::punctuation },
  { 0x16EB, 0x16ED, vank::parse::CharClass::punctuation },
  { 0x1944, 0x1945, vank::parse::CharClass::punctuation },

  { 0x1369, 0x137C, vank::parse::CharClass::digit },
  { 0x0660, 0x0669, vank::parse::CharClass::digit },
  { 0x06F0, 0x06F9, vank::parse::CharClass::digit },
  { 0x1040, 0x1049, vank::parse::CharClass::digit },
  { 0x0AE6, 0x0AEF, vank::parse::CharClass::digit },
  { 0x0D66, 0x0D6F, vank::parse::CharClass::digit },
  { 0x1810, 0x1819, vank::parse::CharClass::digit },
  { 0x0BE6, 0x0BF2, vank::parse::CharClass::digit },
  { 0x1946, 0x194F, vank::parse::CharClass::digit },
  { 0x09E6, 0x09EF, vank::parse::CharClass::digit },
  { 0x09F4, 0x09F9, vank::parse::CharClass::digit },
  { 0x104A0, 0x104A9, vank::parse::CharClass::digit },
  { 0x0ED0, 0x0ED9, vank::parse::CharClass::digit },
  { 0x0F20, 0x0F33, vank::parse::CharClass::digit },
  { 0x0B66, 0x0B6F, vank::parse::CharClass::digit },
  { 0x10107, 0x10133, vank::parse::CharClass::digit },
  { 0x10320, 0x10323, vank::parse::CharClass::digit },
  { 0x00B2, 0x00B3, vank::parse::CharClass::digit },
  { 0x00B9, 0x00B9, vank::parse::CharClass::digit },
  { 0x00BC, 0x00BE, vank::parse::CharClass::digit },
  { 0x1034A, 0x1034A, vank::parse::CharClass::digit },
  { 0x1D7CE, 0x1D7FF, vank::parse::CharClass::digit },
  { 0x0C66, 0x0C6F, vank::parse::CharClass::digit },
  { 0x19D0, 0x19D9, vank::parse::CharClass::digit },
  { 0x0030, 0x0039, vank::parse::CharClass::digit },
  { 0x16EE, 0x16F0, vank::parse::CharClass::digit },
  { 0x0E50, 0x0E59, vank::parse::CharClass::digit },
  { 0x17E0, 0x17E9, vank::parse::CharClass::digit },
  { 0x17F0, 0x17F9, vank::parse::CharClass::digit },
  { 0x3007, 0x3007, vank::parse::CharClass::digit },
  { 0x3021, 0x3029, vank::parse::CharClass::digit },
  { 0x3038, 0x303A, vank::parse::CharClass::digit },
  { 0x3192, 0x3195, vank::parse::CharClass::digit },
  { 0x3220, 0x3229, vank::parse::CharClass::digit },
  { 0x3251, 0x325F, vank::parse::CharClass::digit },
  { 0x3280, 0x3289, vank::parse::CharClass::digit },
  { 0x32B1, 0x32BF, vank::parse::CharClass::digit },
  { 0x0A66, 0x0A6F, vank::parse::CharClass::digit },
  { 0x2070, 0x2070, vank::parse::CharClass::digit },
  { 0x2074, 0x2079, vank::parse::CharClass::digit },
  { 0x2080, 0x2089, vank::parse::CharClass::digit },
  { 0x2153, 0x2183, vank::parse::CharClass::digit },
  { 0x2460, 0x249B, vank::parse::CharClass::digit },
  { 0x24EA, 0x24FF, vank::parse::CharClass::digit },
  { 0x2776, 0x2793, vank::parse::CharClass::digit },
  { 0xFF10, 0xFF19, vank::parse::CharClass::digit },
  { 0x2CFD, 0x2CFD, vank::parse::CharClass::digit },
  { 0x0966, 0x096F, vank::parse::CharClass::digit },
  { 0x103D1, 0x103D5, vank::parse::CharClass::digit },
  { 0x0CE6, 0x0CEF, vank::parse::CharClass::digit },
  { 0x10140, 0x10178, vank::parse::CharClass::digit },
  { 0x1018A, 0x1018A, vank::parse::CharClass::digit },

  { 0x0981, 0x0983, vank::parse::CharClass::letter },
  { 0x09BC, 0x09BC, vank::parse::CharClass::letter },
  { 0x09BE, 0x09C4, vank::parse::CharClass::letter },
  { 0x09C7, 0x09C8, vank::parse::CharClass::letter },
  { 0x09CB, 0x09CD, vank::parse::CharClass::letter },
  { 0x09D7, 0x09D7, vank::parse::CharClass::letter },
  { 0x09E2, 0x09E3, vank::parse::CharClass::letter },
  { 0x10400, 0x1044F, vank::parse::CharClass::letter },
  { 0x10800, 0x10805, vank::parse::CharClass::letter },
  { 0x10808, 0x10808, vank::parse::CharClass::letter },
  { 0x1080A, 0x10835, vank::parse::CharClass::letter },
  { 0x10837, 0x10838, vank::parse::CharClass::letter },
  { 0x1083C, 0x1083C, vank::parse::CharClass::letter },
  { 0x1083F, 0x1083F, vank::parse::CharClass::letter },
  { 0x0300, 0x036F, vank::parse::CharClass::letter },
  { 0x1DC0, 0x1DC3, vank::parse::CharClass::letter },
  { 0x0904, 0x0939, vank::parse::CharClass::letter },
  { 0x093D, 0x093D, vank::parse::CharClass::letter },
  { 0x0950, 0x0950, vank::parse::CharClass::letter },
  { 0x0958, 0x0961, vank::parse::CharClass::letter },
  { 0x097D, 0x097D, vank::parse::CharClass::letter },
  { 0x0610, 0x0615, vank::parse::CharClass::letter },
  { 0x064B, 0x065E, vank::parse::CharClass::letter },
  { 0x0670, 0x0670, vank::parse::CharClass::letter },
  { 0x06D6, 0x06DC, vank::parse::CharClass::letter },
  { 0x06DE, 0x06E4, vank::parse::CharClass::letter },
  { 0x06E7, 0x06E8, vank::parse::CharClass::letter },
  { 0x06EA, 0x06ED, vank::parse::CharClass::letter },
  { 0xFE00, 0xFE0F, vank::parse::CharClass::letter },
  { 0xFE20, 0xFE23, vank::parse::CharClass::letter },
  { 0x05D0, 0x05EA, vank::parse::CharClass::letter },
  { 0x05F0, 0x05F2, vank::parse::CharClass::letter },
  { 0xFB1D, 0xFB1D, vank::parse::CharClass::letter },
  { 0xFB1F, 0xFB28, vank::parse::CharClass::letter },
  { 0xFB2A, 0xFB36, vank::parse::CharClass::letter },
  { 0xFB38, 0xFB3C, vank::parse::CharClass::letter },
  { 0xFB3E, 0xFB3E, vank::parse::CharClass::letter },
  { 0xFB40, 0xFB41, vank::parse::CharClass::letter },
  { 0xFB43, 0xFB44, vank::parse::CharClass::letter },
  { 0xFB46, 0xFB4F, vank::parse::CharClass::letter },
  { 0x0A05, 0x0A0A, vank::parse::CharClass::letter },
  { 0x0A0F, 0x0A10, vank::parse::CharClass::letter },
  { 0x0A13, 0x0A28, vank::parse::CharClass::letter },
  { 0x0A2A, 0x0A30, vank::parse::CharClass::letter },
  { 0x0A32, 0x0A33, vank::parse::CharClass::letter },
  { 0x0A35, 0x0A36, vank::parse::CharClass::letter },
  { 0x0A38, 0x0A39, vank::parse::CharClass::letter },
  { 0x0A59, 0x0A5C, vank::parse::CharClass::letter },
  { 0x0A5E, 0x0A5E, vank::parse::CharClass::letter },
  { 0x0A72, 0x0A74, vank::parse::CharClass::letter },
  { 0x180B, 0x180D, vank::parse::CharClass::letter },
  { 0x18A9, 0x18A9, vank::parse::CharClass::letter },
  { 0x0250, 0x02C1, vank::parse::CharClass::letter },
  { 0x02C6, 0x02D1, vank::parse::CharClass::letter },
  { 0x02E0, 0x02E4, vank::parse::CharClass::letter },
  { 0x02EE, 0x02EE, vank::parse::CharClass::letter },
  { 0x0C01, 0x0C03, vank::parse::CharClass::letter },
  { 0x0C3E, 0x0C44, vank::parse::CharClass::letter },
  { 0x0C46, 0x0C48, vank::parse::CharClass::letter },
  { 0x0C4A, 0x0C4D, vank::parse::CharClass::letter },
  { 0x0C55, 0x0C56, vank::parse::CharClass::letter },
  { 0x2D30, 0x2D65, vank::parse::CharClass::letter },
  { 0x2D6F, 0x2D6F, vank::parse::CharClass::letter },
  { 0x037A, 0x037A, vank::parse::CharClass::letter },
  { 0x0386, 0x0386, vank::parse::CharClass::letter },
  { 0x0388, 0x038A, vank::parse::CharClass::letter },
  { 0x038C, 0x038C, vank::parse::CharClass::letter },
  { 0x038E, 0x03A1, vank::parse::CharClass::letter },
  { 0x03A3, 0x03CE, vank::parse::CharClass::letter },
  { 0x03D0, 0x03F5, vank::parse::CharClass::letter },
  { 0x03F7, 0x03FF, vank::parse::CharClass::letter },
  { 0x302A, 0x302F, vank::parse::CharClass::letter },
  { 0x3099, 0x309A, vank::parse::CharClass::letter },
  { 0x1920, 0x192B, vank::parse::CharClass::letter },
  { 0x1930, 0x193B, vank::parse::CharClass::letter },
  { 0x0E01, 0x0E30, vank::parse::CharClass::letter },
  { 0x0E32, 0x0E33, vank::parse::CharClass::letter },
  { 0x0E40, 0x0E46, vank::parse::CharClass::letter },
  { 0x1780, 0x17B3, vank::parse::CharClass::letter },
  { 0x17D7, 0x17D7, vank::parse::CharClass::letter },
  { 0x17DC, 0x17DC, vank::parse::CharClass::letter },
  { 0x07A6, 0x07B0, vank::parse::CharClass::letter },
  { 0x10380, 0x1039D, vank::parse::CharClass::letter },
  { 0x0483, 0x0486, vank::parse::CharClass::letter },
  { 0x0488, 0x0489, vank::parse::CharClass::letter },
  { 0x17B6, 0x17D3, vank::parse::CharClass::letter },
  { 0x17DD, 0x17DD, vank::parse::CharClass::letter },
  { 0x0E31, 0x0E31, vank::parse::CharClass::letter },
  { 0x0E34, 0x0E3A, vank::parse::CharClass::letter },
  { 0x0E47, 0x0E4E, vank::parse::CharClass::letter },
  { 0x3005, 0x3006, vank::parse::CharClass::letter },
  { 0x3031, 0x3035, vank::parse::CharClass::letter },
  { 0x303B, 0x303C, vank::parse::CharClass::letter },
  { 0x3041, 0x3096, vank::parse::CharClass::letter },
  { 0x309D, 0x309F, vank::parse::CharClass::letter },
  { 0x30A1, 0x30FA, vank::parse::CharClass::letter },
  { 0x30FC, 0x30FF, vank::parse::CharClass::letter },
  { 0x3105, 0x312C, vank::parse::CharClass::letter },
  { 0x3131, 0x318E, vank::parse::CharClass::letter },
  { 0x31A0, 0x31B7, vank::parse::CharClass::letter },
  { 0x31F0, 0x31FF, vank::parse::CharClass::letter },
  { 0x3400, 0x4DB5, vank::parse::CharClass::letter },
  { 0x4E00, 0x9FBB, vank::parse::CharClass::letter },
  { 0xAC00, 0xD7A3, vank::parse::CharClass::letter },
  { 0xF900, 0xFAD9, vank::parse::CharClass::letter },
  { 0xFF66, 0xFFBE, vank::parse::CharClass::letter },
  { 0xFFC2, 0xFFC7, vank::parse::CharClass::letter },
  { 0xFFCA, 0xFFCF, vank::parse::CharClass::letter },
  { 0xFFD2, 0xFFD7, vank::parse::CharClass::letter },
  { 0xFFDA, 0xFFDC, vank::parse::CharClass::letter },
  { 0x20000, 0x2A6D6, vank::parse::CharClass::letter },
  { 0x2F800, 0x2FA1D, vank::parse::CharClass::letter },
  { 0x0A85, 0x0A8D, vank::parse::CharClass::letter },
  { 0x0A8F, 0x0A91, vank::parse::CharClass::letter },
  { 0x0A93, 0x0AA8, vank::parse::CharClass::letter },
  { 0x0AAA, 0x0AB0, vank::parse::CharClass::letter },
  { 0x0AB2, 0x0AB3, vank::parse::CharClass::letter },
  { 0x0AB5, 0x0AB9, vank::parse::CharClass::letter },
  { 0x0ABD, 0x0ABD, vank::parse::CharClass::letter },
  { 0x0AD0, 0x0AD0, vank::parse::CharClass::letter },
  { 0x0AE0, 0x0AE1, vank::parse::CharClass::letter },
  { 0x0531, 0x0556, vank::parse::CharClass::letter },
  { 0x0559, 0x0559, vank::parse::CharClass::letter },
  { 0x055B, 0x055F, vank::parse::CharClass::letter },
  { 0x0561, 0x0587, vank::parse::CharClass::letter },
  { 0xFB13, 0xFB17, vank::parse::CharClass::letter },
  { 0x0D05, 0x0D0C, vank::parse::CharClass::letter },
  { 0x0D0E, 0x0D10, vank::parse::CharClass::letter },
  { 0x0D12, 0x0D28, vank::parse::CharClass::letter },
  { 0x0D2A, 0x0D39, vank::parse::CharClass::letter },
  { 0x0D60, 0x0D61, vank::parse::CharClass::letter },
  { 0x19B0, 0x19C0, vank::parse::CharClass::letter },
  { 0x19C8, 0x19C9, vank::parse::CharClass::letter },
  { 0x0901, 0x0903, vank::parse::CharClass::letter },
  { 0x093C, 0x093C, vank::parse::CharClass::letter },
  { 0x093E, 0x094D, vank::parse::CharClass::letter },
  { 0x0951, 0x0954, vank::parse::CharClass::letter },
  { 0x0962, 0x0963, vank::parse::CharClass::letter },
  { 0x1681, 0x169A, vank::parse::CharClass::letter },
  { 0x0D85, 0x0D96, vank::parse::CharClass::letter },
  { 0x0D9A, 0x0DB1, vank::parse::CharClass::letter },
  { 0x0DB3, 0x0DBB, vank::parse::CharClass::letter },
  { 0x0DBD, 0x0DBD, vank::parse::CharClass::letter },
  { 0x0DC0, 0x0DC6, vank::parse::CharClass::letter },
  { 0x1820, 0x1877, vank::parse::CharClass::letter },
  { 0x1880, 0x18A8, vank::parse::CharClass::letter },
  { 0x0B83, 0x0B83, vank::parse::CharClass::letter },
  { 0x0B85, 0x0B8A, vank::parse::CharClass::letter },
  { 0x0B8E, 0x0B90, vank::parse::CharClass::letter },
  { 0x0B92, 0x0B95, vank::parse::CharClass::letter },
  { 0x0B99, 0x0B9A, vank::parse::CharClass::letter },
  { 0x0B9C, 0x0B9C, vank::parse::CharClass::letter },
  { 0x0B9E, 0x0B9F, vank::parse::CharClass::letter },
  { 0x0BA3, 0x0BA4, vank::parse::CharClass::letter },
  { 0x0BA8, 0x0BAA, vank::parse::CharClass::letter },
  { 0x0BAE, 0x0BB9, vank::parse::CharClass::letter },
  { 0x1900, 0x191C, vank::parse::CharClass::letter },
  { 0x1100, 0x1159, vank::parse::CharClass::letter },
  { 0x115F, 0x11A2, vank::parse::CharClass::letter },
  { 0x11A8, 0x11F9, vank::parse::CharClass::letter },
  { 0x1F00, 0x1F15, vank::parse::CharClass::letter },
  { 0x1F18, 0x1F1D, vank::parse::CharClass::letter },
  { 0x1F20, 0x1F45, vank::parse::CharClass::letter },
  { 0x1F48, 0x1F4D, vank::parse::CharClass::letter },
  { 0x1F50, 0x1F57, vank::parse::CharClass::letter },
  { 0x1F59, 0x1F59, vank::parse::CharClass::letter },
  { 0x1F5B, 0x1F5B, vank::parse::CharClass::letter },
  { 0x1F5D, 0x1F5D, vank::parse::CharClass::letter },
  { 0x1F5F, 0x1F7D, vank::parse::CharClass::letter },
  { 0x1F80, 0x1FB4, vank::parse::CharClass::letter },
  { 0x1FB6, 0x1FC4, vank::parse::CharClass::letter },
  { 0x1FC6, 0x1FD3, vank::parse::CharClass::letter },
  { 0x1FD6, 0x1FDB, vank::parse::CharClass::letter },
  { 0x1FDD, 0x1FEC, vank::parse::CharClass::letter },
  { 0x1FF2, 0x1FF4, vank::parse::CharClass::letter },
  { 0x1FF6, 0x1FFE, vank::parse::CharClass::letter },
  { 0x1712, 0x1714, vank::parse::CharClass::letter },
  { 0xA000, 0xA48C, vank::parse::CharClass::letter },
  { 0x0591, 0x05B9, vank::parse::CharClass::letter },
  { 0x05BB, 0x05BD, vank::parse::CharClass::letter },
  { 0x05BF, 0x05BF, vank::parse::CharClass::letter },
  { 0x05C1, 0x05C2, vank::parse::CharClass::letter },
  { 0x05C4, 0x05C5, vank::parse::CharClass::letter },
  { 0x05C7, 0x05C7, vank::parse::CharClass::letter },
  { 0xFB1E, 0xFB1E, vank::parse::CharClass::letter },
  { 0x1401, 0x1676, vank::parse::CharClass::letter },
  { 0x1A00, 0x1A16, vank::parse::CharClass::letter },
  { 0x1732, 0x1734, vank::parse::CharClass::letter },
  { 0x1772, 0x1773, vank::parse::CharClass::letter },
  { 0x0985, 0x098C, vank::parse::CharClass::letter },
  { 0x098F, 0x0990, vank::parse::CharClass::letter },
  { 0x0993, 0x09A8, vank::parse::CharClass::letter },
  { 0x09AA, 0x09B0, vank::parse::CharClass::letter },
  { 0x09B2, 0x09B2, vank::parse::CharClass::letter },
  { 0x09B6, 0x09B9, vank::parse::CharClass::letter },
  { 0x09BD, 0x09BD, vank::parse::CharClass::letter },
  { 0x09CE, 0x09CE, vank::parse::CharClass::letter },
  { 0x09DC, 0x09DD, vank::parse::CharClass::letter },
  { 0x09DF, 0x09E1, vank::parse::CharClass::letter },
  { 0x09F0, 0x09F1, vank::parse::CharClass::letter },
  { 0x10480, 0x1049D, vank::parse::CharClass::letter },
  { 0x0E81, 0x0E82, vank::parse::CharClass::letter },
  { 0x0E84, 0x0E84, vank::parse::CharClass::letter },
  { 0x0E87, 0x0E88, vank::parse::CharClass::letter },
  { 0x0E8A, 0x0E8A, vank::parse::CharClass::letter },
  { 0x0E8D, 0x0E8D, vank::parse::CharClass::letter },
  { 0x0E94, 0x0E97, vank::parse::CharClass::letter },
  { 0x0E99, 0x0E9F, vank::parse::CharClass::letter },
  { 0x0EA1, 0x0EA3, vank::parse::CharClass::letter },
  { 0x0EA5, 0x0EA5, vank::parse::CharClass::letter },
  { 0x0EA7, 0x0EA7, vank::parse::CharClass::letter },
  { 0x0EAA, 0x0EAB, vank::parse::CharClass::letter },
  { 0x0EAD, 0x0EB0, vank::parse::CharClass::letter },
  { 0x0EB2, 0x0EB3, vank::parse::CharClass::letter },
  { 0x0EBD, 0x0EBD, vank::parse::CharClass::letter },
  { 0x0EC0, 0x0EC4, vank::parse::CharClass::letter },
  { 0x0EC6, 0x0EC6, vank::parse::CharClass::letter },
  { 0x0EDC, 0x0EDD, vank::parse::CharClass::letter },
  { 0x0F00, 0x0F00, vank::parse::CharClass::letter },
  { 0x0F40, 0x0F47, vank::parse::CharClass::letter },
  { 0x0F49, 0x0F6A, vank::parse::CharClass::letter },
  { 0x0F88, 0x0F8B, vank::parse::CharClass::letter },
  { 0x0A81, 0x0A83, vank::parse::CharClass::letter },
  { 0x0ABC, 0x0ABC, vank::parse::CharClass::letter },
  { 0x0ABE, 0x0AC5, vank::parse::CharClass::letter },
  { 0x0AC7, 0x0AC9, vank::parse::CharClass::letter },
  { 0x0ACB, 0x0ACD, vank::parse::CharClass::letter },
  { 0x0AE2, 0x0AE3, vank::parse::CharClass::letter },
  { 0x0B05, 0x0B0C, vank::parse::CharClass::letter },
  { 0x0B13, 0x0B28, vank::parse::CharClass::letter },
  { 0x0B2A, 0x0B30, vank::parse::CharClass::letter },
  { 0x0B32, 0x0B33, vank::parse::CharClass::letter },
  { 0x0B35, 0x0B39, vank::parse::CharClass::letter },
  { 0x0B3D, 0x0B3D, vank::parse::CharClass::letter },
  { 0x0B5C, 0x0B5D, vank::parse::CharClass::letter },
  { 0x0B5F, 0x0B61, vank::parse::CharClass::letter },
  { 0x0B71, 0x0B71, vank::parse::CharClass::letter },
  { 0x1752, 0x1753, vank::parse::CharClass::letter },
  { 0x0780, 0x07A5, vank::parse::CharClass::letter },
  { 0x07B1, 0x07B1, vank::parse::CharClass::letter },
  { 0x10300, 0x1031E, vank::parse::CharClass::letter },
  { 0x0710, 0x0710, vank::parse::CharClass::letter },
  { 0x0712, 0x072F, vank::parse::CharClass::letter },
  { 0x1760, 0x176C, vank::parse::CharClass::letter },
  { 0x176E, 0x1770, vank::parse::CharClass::letter },
  { 0x0B0F, 0x0B10, vank::parse::CharClass::letter },
  { 0x1200, 0x1248, vank::parse::CharClass::letter },
  { 0x124A, 0x124D, vank::parse::CharClass::letter },
  { 0x1250, 0x1256, vank::parse::CharClass::letter },
  { 0x1258, 0x1258, vank::parse::CharClass::letter },
  { 0x125A, 0x125D, vank::parse::CharClass::letter },
  { 0x1260, 0x1288, vank::parse::CharClass::letter },
  { 0x128A, 0x128D, vank::parse::CharClass::letter },
  { 0x1290, 0x12B0, vank::parse::CharClass::letter },
  { 0x12B2, 0x12B5, vank::parse::CharClass::letter },
  { 0x12B8, 0x12BE, vank::parse::CharClass::letter },
  { 0x12C0, 0x12C0, vank::parse::CharClass::letter },
  { 0x12C2, 0x12C5, vank::parse::CharClass::letter },
  { 0x12C8, 0x12D6, vank::parse::CharClass::letter },
  { 0x12D8, 0x1310, vank::parse::CharClass::letter },
  { 0x1312, 0x1315, vank::parse::CharClass::letter },
  { 0x1318, 0x135A, vank::parse::CharClass::letter },
  { 0x1380, 0x138F, vank::parse::CharClass::letter },
  { 0x2D80, 0x2D96, vank::parse::CharClass::letter },
  { 0x2DA0, 0x2DA6, vank::parse::CharClass::letter },
  { 0x2DA8, 0x2DAE, vank::parse::CharClass::letter },
  { 0x2DB0, 0x2DB6, vank::parse::CharClass::letter },
  { 0x2DB8, 0x2DBE, vank::parse::CharClass::letter },
  { 0x2DC0, 0x2DC6, vank::parse::CharClass::letter },
  { 0x2DC8, 0x2DCE, vank::parse::CharClass::letter },
  { 0x2DD0, 0x2DD6, vank::parse::CharClass::letter },
  { 0x2DD8, 0x2DDE, vank::parse::CharClass::letter },
  { 0x1E00, 0x1E9B, vank::parse::CharClass::letter },
  { 0x1EA0, 0x1EF9, vank::parse::CharClass::letter },
  { 0x1950, 0x196D, vank::parse::CharClass::letter },
  { 0x1970, 0x1974, vank::parse::CharClass::letter },
  { 0x0EB1, 0x0EB1, vank::parse::CharClass::letter },
  { 0x0EB4, 0x0EB9, vank::parse::CharClass::letter },
  { 0x0EBB, 0x0EBC, vank::parse::CharClass::letter },
  { 0x0EC8, 0x0ECD, vank::parse::CharClass::letter },
  { 0x1D00, 0x1DBF, vank::parse::CharClass::letter },
  { 0x0621, 0x063A, vank::parse::CharClass::letter },
  { 0x0640, 0x064A, vank::parse::CharClass::letter },
  { 0x066E, 0x066F, vank::parse::CharClass::letter },
  { 0x0671, 0x06D3, vank::parse::CharClass::letter },
  { 0x06D5, 0x06D5, vank::parse::CharClass::letter },
  { 0x06E5, 0x06E6, vank::parse::CharClass::letter },
  { 0x06EE, 0x06EF, vank::parse::CharClass::letter },
  { 0x06FA, 0x06FC, vank::parse::CharClass::letter },
  { 0x06FF, 0x06FF, vank::parse::CharClass::letter },
  { 0x0750, 0x076D, vank::parse::CharClass::letter },
  { 0xFB50, 0xFBB1, vank::parse::CharClass::letter },
  { 0xFBD3, 0xFD3D, vank::parse::CharClass::letter },
  { 0xFD50, 0xFD8F, vank::parse::CharClass::letter },
  { 0xFD92, 0xFDC7, vank::parse::CharClass::letter },
  { 0xFDF0, 0xFDFB, vank::parse::CharClass::letter },
  { 0xFE70, 0xFE74, vank::parse::CharClass::letter },
  { 0xFE76, 0xFEFC, vank::parse::CharClass::letter },
  { 0x0C82, 0x0C83, vank::parse::CharClass::letter },
  { 0x0CBC, 0x0CBC, vank::parse::CharClass::letter },
  { 0x0CBE, 0x0CC4, vank::parse::CharClass::letter },
  { 0x0CC6, 0x0CC8, vank::parse::CharClass::letter },
  { 0x0CCA, 0x0CCD, vank::parse::CharClass::letter },
  { 0x0CD5, 0x0CD6, vank::parse::CharClass::letter },
  { 0x0B82, 0x0B82, vank::parse::CharClass::letter },
  { 0x0BBE, 0x0BC2, vank::parse::CharClass::letter },
  { 0x0BC6, 0x0BC8, vank::parse::CharClass::letter },
  { 0x0BCA, 0x0BCD, vank::parse::CharClass::letter },
  { 0x0BD7, 0x0BD7, vank::parse::CharClass::letter },
  { 0x10A0, 0x10C5, vank::parse::CharClass::letter },
  { 0x10D0, 0x10FA, vank::parse::CharClass::letter },
  { 0x10FC, 0x10FC, vank::parse::CharClass::letter },
  { 0x2D00, 0x2D25, vank::parse::CharClass::letter },
  { 0x0F18, 0x0F19, vank::parse::CharClass::letter },
  { 0x0F35, 0x0F35, vank::parse::CharClass::letter },
  { 0x0F37, 0x0F37, vank::parse::CharClass::letter },
  { 0x0F39, 0x0F39, vank::parse::CharClass::letter },
  { 0x0F3E, 0x0F3F, vank::parse::CharClass::letter },
  { 0x0F71, 0x0F84, vank::parse::CharClass::letter },
  { 0x0F86, 0x0F87, vank::parse::CharClass::letter },
  { 0x0F90, 0x0F97, vank::parse::CharClass::letter },
  { 0x0F99, 0x0FBC, vank::parse::CharClass::letter },
  { 0x0FC6, 0x0FC6, vank::parse::CharClass::letter },
  { 0xA800, 0xA801, vank::parse::CharClass::letter },
  { 0xA803, 0xA805, vank::parse::CharClass::letter },
  { 0xA807, 0xA80A, vank::parse::CharClass::letter },
  { 0xA80C, 0xA822, vank::parse::CharClass::letter },
  { 0x1000, 0x1021, vank::parse::CharClass::letter },
  { 0x1023, 0x1027, vank::parse::CharClass::letter },
  { 0x1029, 0x102A, vank::parse::CharClass::letter },
  { 0x1050, 0x1055, vank::parse::CharClass::letter },
  { 0x20D0, 0x20EB, vank::parse::CharClass::letter },
  { 0x00AA, 0x00AA, vank::parse::CharClass::letter },
  { 0x00B5, 0x00B5, vank::parse::CharClass::letter },
  { 0x00BA, 0x00BA, vank::parse::CharClass::letter },
  { 0x00C0, 0x00D6, vank::parse::CharClass::letter },
  { 0x00D8, 0x00F6, vank::parse::CharClass::letter },
  { 0x00F8, 0x0241, vank::parse::CharClass::letter },
  { 0x2071, 0x2071, vank::parse::CharClass::letter },
  { 0x207F, 0x207F, vank::parse::CharClass::letter },
  { 0x2090, 0x2094, vank::parse::CharClass::letter },
  { 0xFB00, 0xFB06, vank::parse::CharClass::letter },
  { 0xFF21, 0xFF3A, vank::parse::CharClass::letter },
  { 0xFF41, 0xFF5A, vank::parse::CharClass::letter },
  { 0x0711, 0x0711, vank::parse::CharClass::letter },
  { 0x0730, 0x074A, vank::parse::CharClass::letter },
  { 0x074D, 0x074F, vank::parse::CharClass::letter },
  { 0x10330, 0x10349, vank::parse::CharClass::letter },
  { 0x10000, 0x1000B, vank::parse::CharClass::letter },
  { 0x1000D, 0x10026, vank::parse::CharClass::letter },
  { 0x10028, 0x1003A, vank::parse::CharClass::letter },
  { 0x1003C, 0x1003D, vank::parse::CharClass::letter },
  { 0x1003F, 0x1004D, vank::parse::CharClass::letter },
  { 0x10050, 0x1005D, vank::parse::CharClass::letter },
  { 0x10080, 0x100FA, vank::parse::CharClass::letter },
  { 0x1D400, 0x1D454, vank::parse::CharClass::letter },
  { 0x1D456, 0x1D49C, vank::parse::CharClass::letter },
  { 0x1D49E, 0x1D49F, vank::parse::CharClass::letter },
  { 0x1D4A2, 0x1D4A2, vank::parse::CharClass::letter },
  { 0x1D4A5, 0x1D4A6, vank::parse::CharClass::letter },
  { 0x1D4A9, 0x1D4AC, vank::parse::CharClass::letter },
  { 0x1D4AE, 0x1D4B9, vank::parse::CharClass::letter },
  { 0x1D4BB, 0x1D4BB, vank::parse::CharClass::letter },
  { 0x1D4BD, 0x1D4C3, vank::parse::CharClass::letter },
  { 0x1D4C5, 0x1D505, vank::parse::CharClass::letter },
  { 0x1D507, 0x1D50A, vank::parse::CharClass::letter },
  { 0x1D50D, 0x1D514, vank::parse::CharClass::letter },
  { 0x1D516, 0x1D51C, vank::parse::CharClass::letter },
  { 0x1D51E, 0x1D539, vank::parse::CharClass::letter },
  { 0x1D53B, 0x1D53E, vank::parse::CharClass::letter },
  { 0x1D540, 0x1D544, vank::parse::CharClass::letter },
  { 0x1D546, 0x1D546, vank::parse::CharClass::letter },
  { 0x1D54A, 0x1D550, vank::parse::CharClass::letter },
  { 0x1D552, 0x1D6A5, vank::parse::CharClass::letter },
  { 0x1D6A8, 0x1D6C0, vank::parse::CharClass::letter },
  { 0x1D6C2, 0x1D6DA, vank::parse::CharClass::letter },
  { 0x1D6DC, 0x1D6FA, vank::parse::CharClass::letter },
  { 0x1D6FC, 0x1D714, vank::parse::CharClass::letter },
  { 0x1D716, 0x1D734, vank::parse::CharClass::letter },
  { 0x1D736, 0x1D74E, vank::parse::CharClass::letter },
  { 0x1D750, 0x1D76E, vank::parse::CharClass::letter },
  { 0x1D770, 0x1D788, vank::parse::CharClass::letter },
  { 0x1D78A, 0x1D7A8, vank::parse::CharClass::letter },
  { 0x1D7AA, 0x1D7C2, vank::parse::CharClass::letter },
  { 0x1D7C4, 0x1D7C9, vank::parse::CharClass::letter },
  { 0x10A00, 0x10A00, vank::parse::CharClass::letter },
  { 0x10A10, 0x10A13, vank::parse::CharClass::letter },
  { 0x10A15, 0x10A17, vank::parse::CharClass::letter },
  { 0x10A19, 0x10A33, vank::parse::CharClass::letter },
  { 0x10A01, 0x10A03, vank::parse::CharClass::letter },
  { 0x10A05, 0x10A06, vank::parse::CharClass::letter },
  { 0x10A0C, 0x10A0F, vank::parse::CharClass::letter },
  { 0x10A38, 0x10A3B, vank::parse::CharClass::letter },
  { 0x10A3F, 0x10A47, vank::parse::CharClass::letter },
  { 0x0D82, 0x0D83, vank::parse::CharClass::letter },
  { 0x0DCA, 0x0DCA, vank::parse::CharClass::letter },
  { 0x0DCF, 0x0DD4, vank::parse::CharClass::letter },
  { 0x0DD6, 0x0DD6, vank::parse::CharClass::letter },
  { 0x0DD8, 0x0DDF, vank::parse::CharClass::letter },
  { 0x0DF2, 0x0DF3, vank::parse::CharClass::letter },
  { 0x2102, 0x2102, vank::parse::CharClass::letter },
  { 0x2107, 0x2107, vank::parse::CharClass::letter },
  { 0x210A, 0x2113, vank::parse::CharClass::letter },
  { 0x2115, 0x2115, vank::parse::CharClass::letter },
  { 0x2119, 0x211D, vank::parse::CharClass::letter },
  { 0x2124, 0x2124, vank::parse::CharClass::letter },
  { 0x2126, 0x2126, vank::parse::CharClass::letter },
  { 0x2128, 0x2128, vank::parse::CharClass::letter },
  { 0x212A, 0x212D, vank::parse::CharClass::letter },
  { 0x212F, 0x2131, vank::parse::CharClass::letter },
  { 0x2133, 0x2139, vank::parse::CharClass::letter },
  { 0x213C, 0x213F, vank::parse::CharClass::letter },
  { 0x2145, 0x2149, vank::parse::CharClass::letter },
  { 0x1720, 0x1731, vank::parse::CharClass::letter },
  { 0x2C80, 0x2CE4, vank::parse::CharClass::letter },
  { 0x1740, 0x1751, vank::parse::CharClass::letter },
  { 0x2C00, 0x2C2E, vank::parse::CharClass::letter },
  { 0x2C30, 0x2C5E, vank::parse::CharClass::letter },
  { 0x0A01, 0x0A03, vank::parse::CharClass::letter },
  { 0x0A3C, 0x0A3C, vank::parse::CharClass::letter },
  { 0x0A3E, 0x0A42, vank::parse::CharClass::letter },
  { 0x0A47, 0x0A48, vank::parse::CharClass::letter },
  { 0x0A4B, 0x0A4D, vank::parse::CharClass::letter },
  { 0x0A70, 0x0A71, vank::parse::CharClass::letter },
  { 0x13A0, 0x13F4, vank::parse::CharClass::letter },
  { 0x10450, 0x1047F, vank::parse::CharClass::letter },
  { 0x0B01, 0x0B03, vank::parse::CharClass::letter },
  { 0x0B3C, 0x0B3C, vank::parse::CharClass::letter },
  { 0x0B3E, 0x0B43, vank::parse::CharClass::letter },
  { 0x0B47, 0x0B48, vank::parse::CharClass::letter },
  { 0x0B4B, 0x0B4D, vank::parse::CharClass::letter },
  { 0x0B56, 0x0B57, vank::parse::CharClass::letter },
  { 0x0C85, 0x0C8C, vank::parse::CharClass::letter },
  { 0x0C8E, 0x0C90, vank::parse::CharClass::letter },
  { 0x0C92, 0x0CA8, vank::parse::CharClass::letter },
  { 0x0CAA, 0x0CB3, vank::parse::CharClass::letter },
  { 0x0CB5, 0x0CB9, vank::parse::CharClass::letter },
  { 0x0CBD, 0x0CBD, vank::parse::CharClass::letter },
  { 0x0CDE, 0x0CDE, vank::parse::CharClass::letter },
  { 0x0CE0, 0x0CE1, vank::parse::CharClass::letter },
  { 0x103A0, 0x103C3, vank::parse::CharClass::letter },
  { 0x103C8, 0x103D0, vank::parse::CharClass::letter },
  { 0x135F, 0x135F, vank::parse::CharClass::letter },
  { 0x0D02, 0x0D03, vank::parse::CharClass::letter },
  { 0x0D3E, 0x0D43, vank::parse::CharClass::letter },
  { 0x0D46, 0x0D48, vank::parse::CharClass::letter },
  { 0x0D4A, 0x0D4D, vank::parse::CharClass::letter },
  { 0x0D57, 0x0D57, vank::parse::CharClass::letter },
  { 0x0C05, 0x0C0C, vank::parse::CharClass::letter },
  { 0x0C0E, 0x0C10, vank::parse::CharClass::letter },
  { 0x0C12, 0x0C28, vank::parse::CharClass::letter },
  { 0x0C2A, 0x0C33, vank::parse::CharClass::letter },
  { 0x0C35, 0x0C39, vank::parse::CharClass::letter },
  { 0x0C60, 0x0C61, vank::parse::CharClass::letter },
  { 0x1700, 0x170C, vank::parse::CharClass::letter },
  { 0x170E, 0x1711, vank::parse::CharClass::letter },
  { 0x0400, 0x0481, vank::parse::CharClass::letter },
  { 0x048A, 0x04CE, vank::parse::CharClass::letter },
  { 0x04D0, 0x04F9, vank::parse::CharClass::letter },
  { 0x0500, 0x050F, vank::parse::CharClass::letter },
  { 0x1980, 0x19A9, vank::parse::CharClass::letter },
  { 0x19C1, 0x19C7, vank::parse::CharClass::letter },
  { 0x102C, 0x1032, vank::parse::CharClass::letter },
  { 0x1036, 0x1039, vank::parse::CharClass::letter },
  { 0x1056, 0x1059, vank::parse::CharClass::letter },
  { 0xA802, 0xA802, vank::parse::CharClass::letter },
  { 0xA806, 0xA806, vank::parse::CharClass::letter },
  { 0xA80B, 0xA80B, vank::parse::CharClass::letter },
  { 0xA823, 0xA827, vank::parse::CharClass::letter },
  { 0x1A17, 0x1A1B, vank::parse::CharClass::letter },
  { 0x0041, 0x005A, vank::parse::CharClass::letter },
  { 0x0061, 0x007A, vank::parse::CharClass::letter },
  { 0x16A0, 0x16EA, vank::parse::CharClass::letter },

  { 0x1D100, 0x1D126, vank::parse::CharClass::symbol },
  { 0x1D12A, 0x1D1DD, vank::parse::CharClass::symbol },
  { 0x1D300, 0x1D356, vank::parse::CharClass::symbol },
  { 0x10179, 0x10189, vank::parse::CharClass::symbol },
  { 0x002B, 0x002B, vank::parse::CharClass::symbol },
  { 0x003C, 0x003E, vank::parse::CharClass::symbol },
  { 0x007C, 0x007C, vank::parse::CharClass::symbol },
  { 0x1FED, 0x1FEF, vank::parse::CharClass::symbol },
  { 0x1940, 0x1940, vank::parse::CharClass::symbol },
  { 0x09FA, 0x09FA, vank::parse::CharClass::symbol },
  { 0x0B70, 0x0B70, vank::parse::CharClass::symbol },
  { 0x2CE5, 0x2CEA, vank::parse::CharClass::symbol },
  { 0x10102, 0x10102, vank::parse::CharClass::symbol },
  { 0x10137, 0x1013F, vank::parse::CharClass::symbol },
  { 0x1D6C1, 0x1D6C1, vank::parse::CharClass::symbol },
  { 0x1D6DB, 0x1D6DB, vank::parse::CharClass::symbol },
  { 0x1D6FB, 0x1D6FB, vank::parse::CharClass::symbol },
  { 0x1D715, 0x1D715, vank::parse::CharClass::symbol },
  { 0x1D735, 0x1D735, vank::parse::CharClass::symbol },
  { 0x1D74F, 0x1D74F, vank::parse::CharClass::symbol },
  { 0x1D76F, 0x1D76F, vank::parse::CharClass::symbol },
  { 0x1D789, 0x1D789, vank::parse::CharClass::symbol },
  { 0x1D7A9, 0x1D7A9, vank::parse::CharClass::symbol },
  { 0x1D7C3, 0x1D7C3, vank::parse::CharClass::symbol },
  { 0xA828, 0xA82B, vank::parse::CharClass::symbol },
  { 0x060E, 0x060F, vank::parse::CharClass::symbol },
  { 0x06E9, 0x06E9, vank::parse::CharClass::symbol },
  { 0x06FD, 0x06FE, vank::parse::CharClass::symbol },
  { 0xFDFD, 0xFDFD, vank::parse::CharClass::symbol },
  { 0xFE62, 0xFE62, vank::parse::CharClass::symbol },
  { 0xFE64, 0xFE66, vank::parse::CharClass::symbol },
  { 0x2016, 0x2017, vank::parse::CharClass::symbol },
  { 0x2020, 0x2025, vank::parse::CharClass::symbol },
  { 0x2032, 0x2038, vank::parse::CharClass::symbol },
  { 0x203B, 0x203B, vank::parse::CharClass::symbol },
  { 0x2041, 0x2043, vank::parse::CharClass::symbol },
  { 0x204A, 0x204E, vank::parse::CharClass::symbol },
  { 0x2050, 0x2052, vank::parse::CharClass::symbol },
  { 0x2055, 0x205E, vank::parse::CharClass::symbol },
  { 0x207A, 0x207C, vank::parse::CharClass::symbol },
  { 0x208A, 0x208C, vank::parse::CharClass::symbol },
  { 0x2100, 0x2101, vank::parse::CharClass::symbol },
  { 0x2103, 0x2106, vank::parse::CharClass::symbol },
  { 0x2108, 0x2109, vank::parse::CharClass::symbol },
  { 0x2114, 0x2114, vank::parse::CharClass::symbol },
  { 0x2116, 0x2118, vank::parse::CharClass::symbol },
  { 0x211E, 0x2123, vank::parse::CharClass::symbol },
  { 0x2125, 0x2125, vank::parse::CharClass::symbol },
  { 0x2127, 0x2127, vank::parse::CharClass::symbol },
  { 0x2129, 0x2129, vank::parse::CharClass::symbol },
  { 0x212E, 0x212E, vank::parse::CharClass::symbol },
  { 0x2132, 0x2132, vank::parse::CharClass::symbol },
  { 0x213A, 0x213B, vank::parse::CharClass::symbol },
  { 0x2140, 0x2144, vank::parse::CharClass::symbol },
  { 0x214A, 0x214C, vank::parse::CharClass::symbol },
  { 0x2190, 0x23DB, vank::parse::CharClass::symbol },
  { 0x2400, 0x2426, vank::parse::CharClass::symbol },
  { 0x2440, 0x244A, vank::parse::CharClass::symbol },
  { 0x249C, 0x24E9, vank::parse::CharClass::symbol },
  { 0x2500, 0x269C, vank::parse::CharClass::symbol },
  { 0x26A0, 0x26B1, vank::parse::CharClass::symbol },
  { 0x2701, 0x2704, vank::parse::CharClass::symbol },
  { 0x2706, 0x2709, vank::parse::CharClass::symbol },
  { 0x270C, 0x2727, vank::parse::CharClass::symbol },
  { 0x2729, 0x274B, vank::parse::CharClass::symbol },
  { 0x274D, 0x274D, vank::parse::CharClass::symbol },
  { 0x274F, 0x2752, vank::parse::CharClass::symbol },
  { 0x2756, 0x2756, vank::parse::CharClass::symbol },
  { 0x2758, 0x275E, vank::parse::CharClass::symbol },
  { 0x2761, 0x2767, vank::parse::CharClass::symbol },
  { 0x2794, 0x2794, vank::parse::CharClass::symbol },
  { 0x2798, 0x27AF, vank::parse::CharClass::symbol },
  { 0x27B1, 0x27BE, vank::parse::CharClass::symbol },
  { 0x27C0, 0x27C4, vank::parse::CharClass::symbol },
  { 0x27D0, 0x27E5, vank::parse::CharClass::symbol },
  { 0x27F0, 0x2982, vank::parse::CharClass::symbol },
  { 0x2999, 0x29D7, vank::parse::CharClass::symbol },
  { 0x29DC, 0x29FB, vank::parse::CharClass::symbol },
  { 0x29FE, 0x2B13, vank::parse::CharClass::symbol },
  { 0xFF0B, 0xFF0B, vank::parse::CharClass::symbol },
  { 0xFF1C, 0xFF1E, vank::parse::CharClass::symbol },
  { 0x2E80, 0x2E99, vank::parse::CharClass::symbol },
  { 0x2E9B, 0x2EF3, vank::parse::CharClass::symbol },
  { 0x2F00, 0x2FD5, vank::parse::CharClass::symbol },
  { 0x2FF0, 0x2FFB, vank::parse::CharClass::symbol },
  { 0x3004, 0x3004, vank::parse::CharClass::symbol },
  { 0x3012, 0x3013, vank::parse::CharClass::symbol },
  { 0x3020, 0x3020, vank::parse::CharClass::symbol },
  { 0x3036, 0x3037, vank::parse::CharClass::symbol },
  { 0x303E, 0x303F, vank::parse::CharClass::symbol },
  { 0x309B, 0x309C, vank::parse::CharClass::symbol },
  { 0x3200, 0x321E, vank::parse::CharClass::symbol },
  { 0x322A, 0x3243, vank::parse::CharClass::symbol },
  { 0x32C0, 0x32FE, vank::parse::CharClass::symbol },
  { 0x3300, 0x33FF, vank::parse::CharClass::symbol },
  { 0xA700, 0xA716, vank::parse::CharClass::symbol },
  { 0xFFE2, 0xFFE4, vank::parse::CharClass::symbol },
  { 0xFFE8, 0xFFEE, vank::parse::CharClass::symbol },
  { 0xFFFC, 0xFFFD, vank::parse::CharClass::symbol },
  { 0x1360, 0x1360, vank::parse::CharClass::symbol },
  { 0x1390, 0x1399, vank::parse::CharClass::symbol },
  { 0x0F01, 0x0F03, vank::parse::CharClass::symbol },
  { 0x0F13, 0x0F17, vank::parse::CharClass::symbol },
  { 0x0F1A, 0x0F1F, vank::parse::CharClass::symbol },
  { 0x0F34, 0x0F34, vank::parse::CharClass::symbol },
  { 0x0F36, 0x0F36, vank::parse::CharClass::symbol },
  { 0x0F38, 0x0F38, vank::parse::CharClass::symbol },
  { 0x0FBE, 0x0FC5, vank::parse::CharClass::symbol },
  { 0x0FC7, 0x0FCC, vank::parse::CharClass::symbol },
  { 0x0FCF, 0x0FCF, vank::parse::CharClass::symbol },
  { 0x1D200, 0x1D245, vank::parse::CharClass::symbol },
  { 0x4DC0, 0x4DFF, vank::parse::CharClass::symbol },
  { 0xFB29, 0xFB29, vank::parse::CharClass::symbol },
  { 0xA490, 0xA4C6, vank::parse::CharClass::symbol },
  { 0x00A6, 0x00A7, vank::parse::CharClass::symbol },
  { 0x00A9, 0x00A9, vank::parse::CharClass::symbol },
  { 0x00AC, 0x00AC, vank::parse::CharClass::symbol },
  { 0x00AE, 0x00AE, vank::parse::CharClass::symbol },
  { 0x00B0, 0x00B1, vank::parse::CharClass::symbol },
  { 0x00B6, 0x00B6, vank::parse::CharClass::symbol },
  { 0x00D7, 0x00D7, vank::parse::CharClass::symbol },
  { 0x00F7, 0x00F7, vank::parse::CharClass::symbol },
  { 0x0482, 0x0482, vank::parse::CharClass::symbol },
  { 0x1D000, 0x1D0F5, vank::parse::CharClass::symbol },
  { 0x3190, 0x3191, vank::parse::CharClass::symbol },
  { 0x3196, 0x319F, vank::parse::CharClass::symbol },
  { 0x31C0, 0x31CF, vank::parse::CharClass::symbol },
  { 0x3260, 0x327F, vank::parse::CharClass::symbol },
  { 0x328A, 0x32B0, vank::parse::CharClass::symbol },
  { 0x19E0, 0x19FF, vank::parse::CharClass::symbol },
  { 0x03F6, 0x03F6, vank::parse::CharClass::symbol },
  { 0x0BF3, 0x0BF8, vank::parse::CharClass::symbol },
  { 0x0BFA, 0x0BFA, vank::parse::CharClass::symbol },
  { 0x3250, 0x3250, vank::parse::CharClass::symbol },

  { 0x09F2, 0x09F3, vank::parse::CharClass::currency },
  { 0x060B, 0x060B, vank::parse::CharClass::currency },
  { 0xFDFC, 0xFDFC, vank::parse::CharClass::currency },
  { 0xFE69, 0xFE69, vank::parse::CharClass::currency },
  { 0xFFE0, 0xFFE1, vank::parse::CharClass::currency },
  { 0xFFE5, 0xFFE6, vank::parse::CharClass::currency },
  { 0x17DB, 0x17DB, vank::parse::CharClass::currency },
  { 0x0E3F, 0x0E3F, vank::parse::CharClass::currency },
  { 0x0AF1, 0x0AF1, vank::parse::CharClass::currency },
  { 0x0BF9, 0x0BF9, vank::parse::CharClass::currency },
  { 0x0024, 0x0024, vank::parse::CharClass::currency },
  { 0x20A0, 0x20B5, vank::parse::CharClass::currency },
  { 0xFF04, 0xFF04, vank::parse::CharClass::currency },
  { 0x00A2, 0x00A5, vank::parse::CharClass::currency },

  { 0x0000, 0x001F, vank::parse::CharClass::control },
  { 0x007F, 0x007F, vank::parse::CharClass::control },
  { 0x070F, 0x070F, vank::parse::CharClass::control },
  { 0x200B, 0x200F, vank::parse::CharClass::control },
  { 0x2027, 0x2027, vank::parse::CharClass::control },
  { 0x202A, 0x202E, vank::parse::CharClass::control },
  { 0x2060, 0x2063, vank::parse::CharClass::control },
  { 0x206A, 0x206F, vank::parse::CharClass::control },
  { 0xFEFF, 0xFEFF, vank::parse::CharClass::control },
  { 0xFFF9, 0xFFFB, vank::parse::CharClass::control },
  { 0xE0001, 0xE0001, vank::parse::CharClass::control },
  { 0xE0020, 0xE007F, vank::parse::CharClass::control },
  { 0xE0100, 0xE01EF, vank::parse::CharClass::control },
  { 0x0080, 0x009F, vank::parse::CharClass::control },
  { 0x00AD, 0x00AD, vank::parse::CharClass::control },
  { 0x0600, 0x0603, vank::parse::CharClass::control },
  { 0x06DD, 0x06DD, vank::parse::CharClass::control },
  { 0x17B4, 0x17B5, vank::parse::CharClass::control },

  { 0x066A, 0x066A, vank::parse::CharClass::percent },
  { 0x2030, 0x2031, vank::parse::CharClass::percent },
  { 0xFF05, 0xFF05, vank::parse::CharClass::percent },
  { 0x0025, 0x0025, vank::parse::CharClass::percent },

  { 0x055A, 0x055A, vank::parse::CharClass::apostrophe },
  { 0x0027, 0x0027, vank::parse::CharClass::apostrophe },
  { 0x2019, 0x2019, vank::parse::CharClass::apostrophe },
  { 0xFF07, 0xFF07, vank::parse::CharClass::apostrophe },
  { 0, 0, 0 }
};

void vank::parse::UTF8Transcoder::_initHT()
{
  if ( u.size() == 0 ) {
    for( unsigned int i=0; intervals[i].start; i++ ) {
      store_interval( u, intervals[i].start, intervals[i].end,
                      intervals[i].cls );
    }
  }
}
