
#ifdef SWIGGO

/* -----------------------------------------------------------------------------
 * protect_pre.i (also see protect_post.i)
 * use to convert GO panics and C++ exceptions into GO errors.
 *
 * since swig says %rename should be called before a class definition,
 * including the protection code is split into two parts with the protect_pre.i
 * included first, followed by including the class interface definitions, and
 * then the protect_post.i definitions (like a sandwich bread protecting meat).
 *
 * notes:
 * 1) exception handling order controls behavior. Meaning, if the wrong
 * handler is reached first, the resulting behavior may not be as expected.
 * use swig exception.i when C++ throws std_exception, see example usage.
 * wrap C++ code that throws LemurException with setEx(method), see various
 * interface files.
 *
 * 2) TODO: %module_id directive == %module_%unique_id from swig would be nice.
 * swig appends %module name and a unique_id (16 hex digits) as object id suffix.
 * search for unique_id in file ~/Work/swig-3.0.12/Source/Modules/go.cxx
 * the unique_id changes when the primary interface module vank.i is modified.
 *
 * the preferred panic protection pattern requires that we %rename class
 * constructors that are wrapped in order to extend the go interface that
 * wraps the C++ class methods with modified signature that converts panic
 * to go errors.
 *
 * any interface class we wrap with the preferred panic protection pattern
 * needs to be edited to adjust unique ids, and the swig command needs to be
 * run again to ensure the generated go file has consistent unique ids.
 *
 * There is no swig feature that helps with this adjustment automatically.
 * as an example, the egrep command below can be used to find such object ids.
 *
 * assuming our wrapped method includes a unique_id reference that is
 * mismatched when compared to the unique_id generated by swig, edit the
 * interface file and run the swig command again.
 *
 * egrep --color '_vank_go_[0123456789abcdef]{16}' vank_go.go
 * extern void _wrap_Swig_free_vank_go_c2859e9b02de2b??(uintptr_t arg1);
 * ...
 * egrep --color '_vank_go_[0123456789abcdef]{16}' *.i
 * IndexEnvironment_post.i:    return (IndexEnvironment)(SwigcptrWrapped_IndexEnvironment(C._wrap_new_Wrapped_IndexEnvironment_vank_go_8e24520e0b567f??()))
 * ...
 *
 * a) C++ code can vector to SWIG_exception defined in swig's go/exception.i,
 * that dispatches to _swig_gopanic defined in go/goruntime.swg, that in turn
 * crosses into the cgo compiler environment (_cgo_panic). The C/C++ panic
 * message returned is of type "const char *".
 *
 * b) C++ object class methods also vector to the LemurExcepion.i handler,
 * a common exception pattern used in Vank. This pattern requires the C++
 * methods that throw an exception be wrapped with "setEx" to dispatch
 * exceptions to SWIG_exception as discussed above.
 *
 * c) Preferred pattern - extend swig go wrapped arbitrary C/C++ class and
 * add error return type to class method signatures. patterned after
 * reference: https://github.com/jsolmon/go-swig-exceptions
 *
 * 3) Protection wrapping sometimes uses a cut/paste of unwrapped method, see
 * MetadataPairVector_post.i for example. When such a wrapped method presents
 * a variadic argument signature, the C type definition may get modified on
 * re-compile (I think this happens only when the set of interfaces change).
 * When this happens, the swig interface type conversion may need to be
 * adjusted. For example, in an earlier compile the C type C.swig_type_232
 * was defined as C.swig_type_129 for the function below:
 *
 *  func NewWrapped_MetadataPairVector(a ...interface{}) Wrapped_MetadataPairVector {
 *  	argc := len(a)
 *  	if argc == 0 {
 *  		return (Wrapped_MetadataPairVector)(SwigcptrMetadataPairVector(C._wrap_new_MetadataPairVector__SWIG_0_vank_go_add17ee78870902e()))
 *  	}
 *  	if argc == 1 {
 *          _swig_i_0 := a[0].(int)
 *  		return (Wrapped_MetadataPairVector)(SwigcptrMetadataPairVector(C._wrap_new_MetadataPairVector__SWIG_1_vank_go_add17ee78870902e(C.swig_type_232(_swig_i_0))))
 *  	}
 *  	panic("No match for overloaded function call")
 *  }
 *
 *
 *
 * ----------------------------------------------------------------------------- */

//
// start of header section
// this section is header material that does not strictly belong here.
//
// when vank.i is modified, the module unique_id is regenerated, so we collect
// %includes here to avoid manual re-fix-up of the unique_id references. this
// scheme will break when swig code generation becomes smarter and recognizes
// nested %includes, but hopefully by that release comes a new feature
// allowing us to append the unique_id pre-expansion.
//

//
// vank distribution version is a #define in the c header below. here we
// store the version string in a global we can reference from GO.
//
%{
#include "vank/vank-platform.h"
const std::string vankVersion(VANK_DISTRIBUTION);
%}
extern const std::string vankVersion;

//
// this is here so we don't pollute vank.i - buildindex is application code
// and not strictly part of the library.
//
/*
%include "VankBuildIndex.i"
%include "SnippetBuilder.i"
*/

//
// end of header section
//



/*
 * rename classes to further extend its wrapped interface to convert exceptions
 * to go errors. Any method that throws exceptions needs to be wrapped this way.
 */

%rename(Wrapped_Parameters) vank::api::Parameters;
%rename(Wrapped_Load) vank::api::Parameters::Load;


%rename(Wrapped_IndexEnvironment) vank::api::IndexEnvironment;
%rename(Wrapped_setDocumentRoot) vank::api::IndexEnvironment::setDocumentRoot;
%rename(Wrapped_setAnchorTextPath) vank::api::IndexEnvironment::setAnchorTextPath;
%rename(Wrapped_addString) vank::api::IndexEnvironment::addString;
%rename(Wrapped_setOffsetMetadataPath) vank::api::IndexEnvironment::setOffsetMetadataPath;
%rename(Wrapped_setOffsetAnnotationsPath) vank::api::IndexEnvironment::setOffsetAnnotationsPath;
%rename(Wrapped_getFileClassSpec) vank::api::IndexEnvironment::getFileClassSpec;
%rename(Wrapped_addFileClass) vank::api::IndexEnvironment::addFileClass;
%rename(Wrapped_deleteDocument) vank::api::IndexEnvironment::deleteDocument;
%rename(Wrapped_setIndexedFields) vank::api::IndexEnvironment::setIndexedFields;
%rename(Wrapped_setNumericField) vank::api::IndexEnvironment::setNumericField;
%rename(Wrapped_setOrdinalField) vank::api::IndexEnvironment::setOrdinalField;
%rename(Wrapped_setParentalField) vank::api::IndexEnvironment::setParentalField;
%rename(Wrapped_setMetadataIndexedFields) vank::api::IndexEnvironment::setMetadataIndexedFields;
%rename(Wrapped_setStopwords) vank::api::IndexEnvironment::setStopwords;
%rename(Wrapped_setStemmer) vank::api::IndexEnvironment::setStemmer;
%rename(Wrapped_setMemory) vank::api::IndexEnvironment::setMemory;
%rename(Wrapped_setMaxDocs) vank::api::IndexEnvironment::setMaxDocs;
%rename(Wrapped_setNormalization) vank::api::IndexEnvironment::setNormalization;
%rename(Wrapped_setStoreDocs) vank::api::IndexEnvironment::setStoreDocs;
%rename(Wrapped_create) vank::api::IndexEnvironment::create;
%rename(Wrapped_open) vank::api::IndexEnvironment::open;
%rename(Wrapped_close) vank::api::IndexEnvironment::close;
%rename(Wrapped_addFile) vank::api::IndexEnvironment::addFile;
%rename(Wrapped_addString) vank::api::IndexEnvironment::addString;
%rename(Wrapped_addParsedDocument) vank::api::IndexEnvironment::addParsedDocument;
%rename(Wrapped_documentsIndexed) vank::api::IndexEnvironment::documentsIndexed;
%rename(Wrapped_documentsSeen) vank::api::IndexEnvironment::documentsSeen;


%rename(Wrapped_deleteFileClassSpec) deleteFileClassSpec;


%rename(Wrapped_MetadataPair) vank::parse::MetadataPair;
// the commented out rename will not work because the reference is a template
// name, and not a class name. So we need to create a wrapped class manually.
// see MetadataPairVector_post.i
//%rename(Wrapped_MetadataPairVector) MetadataPairVector;

/*
%rename(Wrapped_InfoSpace) ride::api::InfoSpace;
%rename(Wrapped_InfoSpace_IsAttached) ride::api::InfoSpace::IsAttached;
%rename(Wrapped_InfoSpace_IsDetached) ride::api::InfoSpace::IsDetached;
%rename(Wrapped_InfoSpace_InitSchema) ride::api::InfoSpace::InitSchema;
%rename(Wrapped_InfoSpace_VerifySchemaPostInit) ride::api::InfoSpace::VerifySchemaPostInit;
%rename(Wrapped_InfoSpace_ReclaimSchema) ride::api::InfoSpace::ReclaimSchema;
%rename(Wrapped_InfoSpace_InitCodecFactory) ride::api::InfoSpace::InitCodecFactory;
%rename(Wrapped_InfoSpace_LookupCodecFactory) ride::api::InfoSpace::LookupCodecFactory;
%rename(Wrapped_InfoSpace_GenROMeta) ride::api::InfoSpace::GenROMeta;
%rename(Wrapped_InfoSpace_GenRWMeta) ride::api::InfoSpace::GenRWMeta;
%rename(Wrapped_InfoSpace_GetWindow) ride::api::InfoSpace::GetWindow;
%rename(Wrapped_InfoSpace_GetWindowEnd) ride::api::InfoSpace::GetWindowEnd;
%rename(Wrapped_InfoSpace_GetDes) ride::api::InfoSpace::GetDes;
%rename(Wrapped_InfoSpace_GetApp) ride::api::InfoSpace::GetApp;
%rename(Wrapped_InfoSpace_GetName) ride::api::InfoSpace::GetName;
%rename(Wrapped_InfoSpace_GetServer) ride::api::InfoSpace::GetServer;
%rename(Wrapped_InfoSpace_GetCommand) ride::api::InfoSpace::GetCommand;
%rename(Wrapped_InfoSpace_GetIndexPath) ride::api::InfoSpace::GetIndexPath;
%rename(Wrapped_InfoSpace_GetCorpusPath) ride::api::InfoSpace::GetCorpusPath;
%rename(Wrapped_InfoSpace_GetCorpusClass) ride::api::InfoSpace::GetCorpusClass;
%rename(Wrapped_InfoSpace_GetBuildParam) ride::api::InfoSpace::GetBuildParam;
%rename(Wrapped_InfoSpace_GetStopParam) ride::api::InfoSpace::GetStopParam;
%rename(Wrapped_InfoSpace_Attach) ride::api::InfoSpace::Attach;
%rename(Wrapped_InfoSpace_Detach) ride::api::InfoSpace::Detach;
%rename(Wrapped_InfoSpace_Destroy) ride::api::InfoSpace::Destroy;
%rename(Wrapped_InfoSpace_AddDoc) ride::api::InfoSpace::AddDoc;
%rename(Wrapped_InfoSpace_AddFile) ride::api::InfoSpace::AddFile;
%rename(Wrapped_InfoSpace_AddString) ride::api::InfoSpace::AddString;
%rename(Wrapped_InfoSpace_DeleteDomain) ride::api::InfoSpace::DeleteDomain;
%rename(Wrapped_InfoSpace_EncodeTime) ride::api::InfoSpace::EncodeTime;
%rename(Wrapped_InfoSpace_DecodeTime) ride::api::InfoSpace::DecodeTime;
%rename(Wrapped_InfoSpace_GetPos) ride::api::InfoSpace::GetPos;
%rename(Wrapped_InfoSpace_SetPos) ride::api::InfoSpace::SetPos;
%rename(Wrapped_InfoSpace_GetMax) ride::api::InfoSpace::GetMax;
%rename(Wrapped_InfoSpace_SetMax) ride::api::InfoSpace::SetMax;
%rename(Wrapped_InfoSpace_CreateCodecSet) ride::api::InfoSpace::CreateCodecSet;
%rename(Wrapped_InfoSpace_DestroyCodecSet) ride::api::InfoSpace::DestroyCodecSet;
%rename(Wrapped_InfoSpace_NFLess) ride::api::InfoSpace::NFLess;
%rename(Wrapped_InfoSpace_NFGreater) ride::api::InfoSpace::NFGreater;
%rename(Wrapped_InfoSpace_NFBetween) ride::api::InfoSpace::NFBetween;
%rename(Wrapped_InfoSpace_NFEquals) ride::api::InfoSpace::NFEquals;
%rename(Wrapped_InfoSpace_Find) ride::api::InfoSpace::Find;
%rename(Wrapped_InfoSpace_DomainExists) ride::api::InfoSpace::DomainExists;
%rename(Wrapped_InfoSpace_GetDomain) ride::api::InfoSpace::GetDomain;
%rename(Wrapped_InfoSpace_GetDomains) ride::api::InfoSpace::GetDomains;
%rename(Wrapped_InfoSpace_FindExtent) ride::api::InfoSpace::FindExtent;

%rename(Wrapped_Domain) ride::api::Domain;
%rename(Wrapped_Domain_ReclaimProps) ride::api::Domain::ReclaimProps;
%rename(Wrapped_Domain_GetFd) ride::api::Domain::GetFd;
%rename(Wrapped_Domain_SetFd) ride::api::Domain::SetFd;
%rename(Wrapped_Domain_GetFds) ride::api::Domain::GetFds;
%rename(Wrapped_Domain_GetMeta) ride::api::Domain::GetMeta;
%rename(Wrapped_Domain_SetMeta) ride::api::Domain::SetMeta;
%rename(Wrapped_Domain_GetROMeta) ride::api::Domain::GetROMeta;
%rename(Wrapped_Domain_GetRWMeta) ride::api::Domain::GetRWMeta;
%rename(Wrapped_Domain_GetCursor) ride::api::Domain::GetCursor;
%rename(Wrapped_Domain_SetCursor) ride::api::Domain::SetCursor;
%rename(Wrapped_Domain_GetField) ride::api::Domain::GetField;
%rename(Wrapped_Domain_SetField) ride::api::Domain::SetField;
%rename(Wrapped_Domain_AddField) ride::api::Domain::AddField;
%rename(Wrapped_Domain_GetFields) ride::api::Domain::GetFields;
%rename(Wrapped_Domain_GetFieldCount) ride::api::Domain::GetFieldCount;
%rename(Wrapped_Domain_PrintTagIndent) ride::api::Domain::PrintTagIndent;
%rename(Wrapped_Domain_PrintTagOpenValue) ride::api::Domain::PrintTagOpenValue;
%rename(Wrapped_Domain_PrintTagClose) ride::api::Domain::PrintTagClose;
%rename(Wrapped_Domain_PrintNode) ride::api::Domain::PrintNode;
%rename(Wrapped_Domain_RenderToString) ride::api::Domain::RenderToString;
%rename(Wrapped_Domain_Save) ride::api::Domain::Save;
%rename(Wrapped_Domain_SaveToFile) ride::api::Domain::SaveToFile;
*/

%rename(Wrapped_StringPair) std::pair< std::string, std::string >;

// %rename must precede declaration
/* %include "Odms.i"
%include "DocIdVector.i"
%include "StringPair.i"
%include "StringPairVector.i"
%include "DomainPtrVector.i"
*/

#endif
